/*
Author: Dominic Nicholson
Description:
	This program creates charts from a json file. It allows the user to graph the given data by custom functions
and filters. The user can have as many graphs and overlaid plots as needed. 
This is implemented in the following manner:
All graphs are redrawn by 'GenerateCharts'. GenerateCharts then calls 'GenerateGraph' to generate each graph individualy
Each graph is independant, and are differentiated by index. GenerateGraph will know which graph to draw by the value of ConfigChart_CurrentIndex.
GenerateGraph calls 'GenerateCategorySelect' to generate the filters window and 'GenerateFunctionEdit' to generate the function window.
It then calls 'ParseData' which returns computed values for the Y component of each datapoint for each filter. Then the results are
used to generate paths and nodes in a graph. Results are computed by running the user-defined function through a custom interpreter.

What's missing:
You can hide filters but not delete them. Add a way to remove a filter without disrupting the color scheme or lettering.
*/

var ConfigChart_hostname = "http://dominicnicholson.no-ip.biz/RegVis/";	//The hostname and base path where the .json is located
var ConfigChart_Height = 500;											//The height of a single chart
var ConfigChart_Width;													//ReadOnly. Assigned at runtime
var ConfigChart_NodeWidth = 80;											//The x-axis distance between each point on a graph
var ConfigChart_NodeRadius = 4;											//Radius of the circle points
var ConfigChart_PathWidth = 3;											//The width of the paths drawn for each chart
var ConfigChart_AxisWidth = 60;											//The width of the X and Y axis label areas
var ConfigChart_TitlebarHeight = 25;									//The height of the titlebar on each chart
var ConfigChart_YDivisions = 10;										//The number of divisions to subdivide the Y axis
var ConfigChart_DiagramYPadding = .1; 									//Padding above and below paths. Valid values are between 0 and 1
var ConfigChart_DiagramXPadding = 30; 									//Padding to the left and right of paths in pixels
var ConfigChart_BorderWidth = 6; 										//Width of the border around each chart
var ConfigChart_OuterMargins = 4;										//Margins for the entire window
var ConfigChart_InnerMargins = 20;										//Spacing between each chart
var ConfigChart_Roundness = 16;											//The roundness of the corners of each chart
var ConfigChart_FunctionWidth = 250;									//The width of the function and filter windows
var ConfigChart_CaretPos;												//The last knows cursor position
var ConfigChart_json;													//Raw data read directly from the json file
var ConfigChart_SymbolTable = [];										//The symbol table used when computing values for a chart's path (cleared after each X-axis datapoint group)

var ConfigChart_Functions = [];											//ConfigChart_Functions[ChartNumber][FunctionNumber] = Data for the specified function including its tokens and jump table 
var ConfigChart_Filters = [];											//ConfigChart_Filters[ChartNumber][FilterNumber] = Describes a filter and which function it is attached to.
var ConfigChart_Settings = [];											//ConfigChart_Settings[ChartNumber] = Mostly UI settings for a specific chart. EX: Whether the function window is toggled.

var ConfigChart_Cont;													//Main container for everything else
var ConfigChart_CurrentIndex = 0; 										//A number indicating the chart we are generating
var ConfigChart_NumCharts = 0;											//The number of charts currently in use

var ConfigChart_AllDepartments;											//Extracted from the json. A set of all departments found.
var ConfigChart_AllCatalogs;											//Extracted from the json. A set of all catalogs found.
var ConfigChart_AllValueKeys;											//Extracted from the json. A set of all value keys found.
var ConfigChart_AllCategories;											//Extracted from the json. A set of all catagories found.

var ConfigChart_KeyPressEvent = null;									//Holds an event handler so we can adjust when to apply syntax highlighting


init();


function loadJSON(callback) {
	//Load the json and send it to the caller
	var xobj = new XMLHttpRequest();
	xobj.overrideMimeType("application/json");

	xobj.open('GET', ConfigChart_hostname + 'ConfigurableChart.json',true);

	xobj.onreadystatechange = function () {
		if (xobj.readyState == 4 && xobj.status == 200) {
			callback(xobj.responseText);
		}
	};
	xobj.send(null);
}
//Parse the data for use in the diagram
function init() {
	loadJSON(function (response) {
		ConfigChart_json = JSON.parse(response);

		//Sort the data properly
		ConfigChart_json.data.sort(function (a,b) {
			if (a.Term < b.Term) return -1;
			if (a.Term == b.Term) {
				if (a.Department < b.Department) return 1;
				if (a.Department == b.Department) {
					if (a.Catalog < b.Catalog) return -1;
					if (a.Catalog == b.Catalog) return 0;
					return 1;
				}
				return -1;
			}
			return 1;
		});

		//Sets of data relavant to filtering ect
		ConfigChart_AllDepartments = new Set();
		ConfigChart_AllCatalogs = new Set();
		ConfigChart_AllCategories = new Set();
		ConfigChart_AllValueKeys = new Set();

		//Extract all of the data for the sets from the json
		ConfigChart_json.data.forEach(function(d,i) {

			if (typeof ConfigChart_json.classes[d.Department + " " + d.Catalog] === 'undefined') {
				console.log(d.Department + " " + d.Catalog + " has not been defined in the 'classes' section of the datafile")
				ConfigChart_json.data.splice(i,i+1);
				return;
			}

			ConfigChart_AllDepartments.add(d.Department);
			ConfigChart_AllCatalogs.add(d.Catalog);

			for (i = 0; i < ConfigChart_json.classes[d.Department + " " + d.Catalog].length; ++i)
				ConfigChart_AllCategories.add(ConfigChart_json.classes[d.Department + " " + d.Catalog][i]);

			for (i = 0; i < d.Values.length; ++i)
				ConfigChart_AllValueKeys.add(d.Values[i].key);
		});

		//Initialize the default function
		ConfigChart_Functions[0] = [];
		ConfigChart_Functions[0][0] = {"func":[],"tokens":[],"jump":[],"ignore":true};
		ConfigChart_json.Function.forEach(function (d) {
			ConfigChart_Functions[0][0].func.push(d);
		});
		Tokenize(0,0);

		//Add the first chart
		AddChart();
	});

}
function GenerateCharts() {
	//Regenerates all charts 


	var scrollpos = window.pageYOffset;	//Needed to maintain scroll when redrawing

	//Save any changes to the function the user may have made. Unless we have just changed filters or functions.
	for (i = 0; i < ConfigChart_NumCharts; ++i) {
		if (Function != null && !ConfigChart_Functions[i].ignore) {
			var Function = d3.select("#EditFunction"+i)[0][0];
			var funcnum = ConfigChart_Filters[i][ConfigChart_Settings[i].FilterNum].func;
			ConfigChart_Functions[i][funcnum].func = Function.innerText.split("\n");
			Tokenize(i,funcnum);
		}
		ConfigChart_Functions[i].ignore = false;	//Don't ignore changes next time
	}

	if (typeof ConfigChart_Cont != 'undefined') ConfigChart_Cont.remove();	//clear all previous charts

	ConfigChart_Cont = d3.select("body").append("div").attr("id","container").style("margin", ConfigChart_OuterMargins);

	//Redraw all charts
	for (ConfigChart_CurrentIndex = 0; ConfigChart_CurrentIndex < ConfigChart_NumCharts; ++ConfigChart_CurrentIndex)
		GenerateGraph();


	//Apply syntax highlighting
	SyntaxHighlighting();

	//Create a button to add a new chart
	ConfigChart_Cont.append('button')
		.attr("class", "mybutton")
		.style("margin", 10)
		.on("click",AddChart)
		.text("Add Chart")
	;

	//Scroll back to the last position we were at
	window.scrollTo(0,scrollpos);
}
	
function GenerateGraph() {
	//Generates a graph for index 'ConfigChart_CurrentIndex'

	//Get an alternating color for the background
	var color;
	if (ConfigChart_CurrentIndex %2) color = "#FAFCFF";
	else color = "#FFFFFF";

	//Get the offset for the current chart. A combination of many independant values.
	var chartpos = ConfigChart_CurrentIndex*(ConfigChart_Height + ConfigChart_AxisWidth + ConfigChart_TitlebarHeight + ConfigChart_BorderWidth*2 + ConfigChart_InnerMargins)
		+ ConfigChart_OuterMargins;

	//Draw the backgound color
	ConfigChart_Cont.append('div').style("width", window.innerWidth)
		.style("height", ConfigChart_Height + ConfigChart_AxisWidth + ConfigChart_TitlebarHeight + ConfigChart_BorderWidth*2 + ConfigChart_InnerMargins)
		.style("position","	absolute")
		.style("top", ConfigChart_OuterMargins - ConfigChart_InnerMargins/2 + ConfigChart_CurrentIndex*(ConfigChart_Height + ConfigChart_AxisWidth + ConfigChart_TitlebarHeight + ConfigChart_BorderWidth*2 + ConfigChart_InnerMargins))
		.style("left", 0)
		.style("z-index", -1)
		.style("background-color", color)
	;

	//Container for the main chart section
	var chart = ConfigChart_Cont.append("table").attr("id","chart").attr("cellpadding", 0)
		.attr("cellspacing", 0)
		.attr("border", ConfigChart_BorderWidth)
		.style("margin-top",0)
		.style("margin-bottom",ConfigChart_InnerMargins)
		.style("border-top-right-radius",ConfigChart_Roundness + 'px')
		.style("border-bottom-right-radius",ConfigChart_Roundness + 'px')
	;

	//Container for the function and filter windows
	var tools = ConfigChart_Cont.append("table").attr("id","tools")
		.attr("cellpadding", 0)
		.style("float","right")
		.style("position", "absolute")
		.style("top", chartpos)
		.style("right", ConfigChart_OuterMargins)
		.attr("cellspacing", 0)
		.attr("border", ConfigChart_BorderWidth)
		.style("border-top-left-radius",ConfigChart_Roundness + 'px')
		.style("border-bottom-left-radius",ConfigChart_Roundness + 'px')
	;

	tools.append("tr").attr("id", "tools"+ConfigChart_CurrentIndex);

	//If both windows are closed don't draw the window border
	if (!ConfigChart_Settings[ConfigChart_CurrentIndex].ShowFilters && !ConfigChart_Settings[ConfigChart_CurrentIndex].ShowFunction) 
		tools.attr("border", 0);

	//Titlebar for the chart
	var titlebar = chart.append("tr").append("td")
		.attr("colspan",2)
		.attr("id", "title"+ConfigChart_CurrentIndex)
		.style("border", 0)
		.style("border-bottom", "solid 1px")
		.style("border-top-right-radius",(ConfigChart_Roundness-ConfigChart_BorderWidth) + 'px')
		.style("background-color","BBCCEE")
		.style("height", ConfigChart_TitlebarHeight)
	;

	//Add the subtile showing which filters are selected
	titlebar.append('div')
		.attr("class","SubTitle")
		.data([ConfigChart_CurrentIndex])
		.style("margin", 0)
		.style("margin-left", 10)
		.style("margin-top", 4)
		.style("float", "left")
		.text(function(d) {

			//Get the current filter that is being shown
			var curfilters = ConfigChart_Filters[ConfigChart_CurrentIndex][ConfigChart_Settings[ConfigChart_CurrentIndex].FilterNum];
			var tmp = "Filter: ";

			//Variable helpful for deciding when to place '&' and when not to.
			var A = curfilters.categories.size > 0;
			var B = curfilters.departments.size > 0;
			var C = curfilters.classes.size > 0;

			//Generate the filter's boolean expression 
			if (A) {
				tmp += "(";
				var i = 0;
				curfilters.categories.forEach(function(d) {
					if (i != 0) tmp += '|' + d;					
					else tmp += d;
					++i;
				});
				tmp += ")";
			}

			if (A & B) tmp += " & ";

			if (B) {
				tmp += "(";
				var i = 0;
				curfilters.departments.forEach(function(d) {
					if (i != 0) tmp += '|' + d;					
					else tmp += d;
					++i;
				});
				tmp += ")";
			}

			if (C & (B | A)) tmp += " & ";

			if (C) {
				tmp += "(";
				var i = 0;
				curfilters.classes.forEach(function(d) {
					if (i != 0) tmp += '|' + d;					
					else tmp += d;
					++i;
				});
				tmp += ")";
			}


			return tmp;
		})
	;
	//If this is not the only chart, aloow it to be deleted
	if (ConfigChart_CurrentIndex != 0) {
		titlebar.append("button")
			.data([ConfigChart_CurrentIndex])
			.attr("class", "mybutton")
			.style("float", "right")
			.style("font-size", 14)
			.style("margin-top", 0)
			.on("click", function (d) {
				RemoveChart(d);
			})
			.text("X")
		;
	}

	var row1 = chart.append("tr").attr("id", "chart"+ConfigChart_CurrentIndex);
	var row2 = chart.append("tr").attr("id", "chartr2i"+ConfigChart_CurrentIndex);

	var selection = row1.append("td").style("border", 0);
	var key = row1.append('td').style("background-color","#DDEEFF").style("vertical-align","top").style("border", 0);
	key.append('div').append("text").html("Key:<br>").attr("class", "SubTitle");

	//Generate the filter and function windows
	GenerateCategorySelect();
	GenerateFunctionEdit();

	//Do the actual computation for each path
	var results = ParseData();	

	//Make sure we don't try drawing nothingness
	var noresults = true;
	results.forEach(function(d) {
		if (d.length != 0) noresults = false;
	});

	//If there are no results, say so clearly
	if (noresults) {
		selection
			.style("height",ConfigChart_Height + ConfigChart_AxisWidth+ConfigChart_BorderWidth)
			.text("No results")
		;
	
		return;
	}

	//Calculate the required width
	ConfigChart_Width = ConfigChart_NodeWidth * (results.terms.length-1) + ConfigChart_DiagramXPadding*2;

	//Create the axis
	var Axis = selection.append("div").attr("id","axis");
	Axis.style("pointer-events", "none");

	var XAxis = row2.append("td").attr("colspan",2);
	XAxis
		.attr("id","xAxis")
		.style("height",ConfigChart_AxisWidth)
		.style("background-color","AADDFF")
		.style("border", 0) 
		.style("border-bottom-right-radius",(ConfigChart_Roundness-ConfigChart_BorderWidth) + 'px')
	;
	var YAxis = Axis.append("table");
	YAxis
		.attr("id","yAxis")
		.style("width",ConfigChart_AxisWidth + ConfigChart_Width)
		.style("height",ConfigChart_Height)
		.attr("cellpadding", 0)
		.attr("cellspacing", 0)
		.attr("border", 0)
	;

	//Y offset for the svg element
	var offset = ConfigChart_OuterMargins + ConfigChart_TitlebarHeight + ConfigChart_BorderWidth 
		+ ConfigChart_CurrentIndex*(ConfigChart_Height + ConfigChart_AxisWidth + ConfigChart_InnerMargins + ConfigChart_BorderWidth*2 + ConfigChart_TitlebarHeight);

	//Holds each path
	var svg = selection.append("svg:svg")
		.style("position","absolute")
		.style("left", ConfigChart_AxisWidth+ConfigChart_BorderWidth + ConfigChart_OuterMargins)
		.style("top",offset)
		.attr("width", ConfigChart_Width)
		.attr("height",ConfigChart_Height)
	;

	var links = svg.append("svg:g").attr("id","links");
	var nodes = svg.append("svg:g").attr("id","nodes");

	//Normalize the results so that there is spacing above and below the diagram
	var difference = results.maxvalue-results.minvalue;
	var normalizedMin = results.minvalue - difference*ConfigChart_DiagramYPadding;
	var normalizedMax = results.maxvalue + difference*ConfigChart_DiagramYPadding;
	var normalizedDiff = normalizedMax - normalizedMin;

	//If there is no difference betweeen max and min values create one artificially
	if (normalizedDiff == 0) {
		normalizedDiff = 2;
		--normalizedMin;
		++normalizedMax;
	}

	//Subdivide the graph vertically 
	for (i = .5; i < ConfigChart_YDivisions; ++i) {
		var percentage = (i/ConfigChart_YDivisions);
		var nodeheight = ConfigChart_Height / ConfigChart_YDivisions;
		var row = YAxis.append("tr");

		var color;
		if (Math.round(i) % 2) color = "#F0F0F0";
		else color = "#FFFFFF";

		row.append("td")
			.style("background-color","#DDEEFF")
			.style("width", ConfigChart_AxisWidth)
			.style("height", nodeheight)
			.style("text-align", "center")
			.style("vertical-align", "middle")
			.text((normalizedMin*percentage + normalizedMax*(1-percentage)).toFixed(1))
		var sel = row.append("td");
		sel.append('div')
			.style("width", ConfigChart_Width)
			.style("height", nodeheight/2)
			.style("background-color","#F0F0F0")
			.style("opacity",".2")
		;
		sel.append('div')
			.style("width", ConfigChart_Width)
			.style("height", nodeheight/2)
			.style("background-color","#FFFFFF")
			.style("opacity",".2")
		;
	}

	//Generate the X-axis labels
	results.terms.forEach(function(d,i) {
		var x = (i)*ConfigChart_NodeWidth + ConfigChart_DiagramXPadding;

		var termname;
		if (d[3] == '3')
			termname = "F " + d.substr(1,2);
		else termname = "Sp " + d.substr(1,2);

		XAxis.append("div")
			.style("left", x+ConfigChart_AxisWidth-ConfigChart_NodeWidth/2 + ConfigChart_BorderWidth + ConfigChart_OuterMargins)
			.style("margin-top",-ConfigChart_AxisWidth/2)
			.style("position", "absolute")
			.style("text-align", "center")
			.style("width", ConfigChart_NodeWidth)

			.html(termname)
		;
	});

	//Initialize the random colors that will be needed
	var colors = d3.scale.category10();
	for (i = 0; i < results.length; ++i) colors(i);

	//Draw the results for each filter
	results.forEach(function(data,graphnum) {

		//Skip this filter if the user has chosen to hide it
		if (ConfigChart_Settings[ConfigChart_CurrentIndex].HiddenFilters.has(graphnum)) return;
		var group = links.append('svg:g');

		//Draw the datapoints in order
		data.sort(function(a,b) {
			if (a.term < b.term) return -1;
			if (a.term == b.term) return 0;
			return 1;
		});

		//Draw the results for filter 'graphnum' as a path with nodes
		data.forEach(function(d,i) {
			
			var termnum = results.terms.indexOf(d.term);
			d.x = (termnum)*ConfigChart_NodeWidth + ConfigChart_DiagramXPadding;
			d.y = ConfigChart_Height - ((d.value-normalizedMin)/normalizedDiff) * ConfigChart_Height;

			//Draw the points of the graph
			nodes.append("circle")
				.data([{"graphnum":graphnum, "chartnum":ConfigChart_CurrentIndex, "d":d}])
				.attr("r", ConfigChart_NodeRadius)
				.attr("cx", d.x)
				.attr("cy", d.y)
				.attr("fill",colors(graphnum))
				.on("click", function (d) {
					//Select this filter if this graph is clicked
					ConfigChart_Settings[d.chartnum].FilterNum = d.graphnum;
					GenerateCharts();
				})
				.append("svg:title")
					.text(
						function(d) {
							//Add a tooltip for the point

							var termname;
							if (d.d.term[3] == '3')
								termname = "Fall " + d.d.term.substr(1,2);
							else termname = "Spring " + d.d.term.substr(1,2);

							var txt = "value: " + d.d.value + ", term: " + termname + "\nComponents:\n";
							d.d.components.forEach(function(d) {
								txt += Math.round(d.value*10)/10 + " from " + d.key + "\n";
							});
							return txt;
						}
					)
			;

			//You can't draw a line with only 1 point so skip the first data point
			if (i != 0 && data.length > 1) {	
				var path = "";

				path += "M " + data[i-1].x + " " + data[i-1].y;	//Start at the previous point
				path += " L " + d.x + " " + d.y;				//End at this datapoint

				var dy = (d.value - data[i-1].value);

				//Draw the path
				group.append("svg:path")
					.attr("d", path)
					.attr("fill", "none")
					.attr("stroke", colors(graphnum))
					.style("stroke-width", ConfigChart_PathWidth)
					.style('opacity',.5)
					.data([{"graphnum":graphnum, "chartnum":ConfigChart_CurrentIndex, 'all': links,'links':group}])
					.on("mouseover", function (d) {
						//On mouseover highlight this path
						d.all.selectAll('path').forEach(function(d) {
							d.forEach(function(d) {d.style['opacity'] = .2;});
						});

						d.links.selectAll('path').forEach(function(d) {
							d.forEach(function(d) {d.style['opacity'] = 1;});
						});
					})
					.on("mouseleave", function (d) {
						//On mouseveave deselect this path
						d.all.selectAll('path').forEach(function(d) {
							d.forEach(function(d) {d.style['opacity'] = .5;});
						});

					})
					.on("click", function (d) {
						//Select this filter if this graph is clicked
						ConfigChart_Settings[d.chartnum].FilterNum = d.graphnum;
						GenerateCharts();
					})
					.append("svg:title")
						.text(
							"dy: " + dy
						)
				;
			}


		});

		//Generate the key at the right side of the chart
		var k = key.append('div')
			.data([{"graphnum":graphnum, "chartnum":ConfigChart_CurrentIndex}])
			.on("click", function (d) {
				ConfigChart_Settings[d.chartnum].FilterNum = d.graphnum;
				GenerateCharts();
			})
			.style("text-align", "center")
			.style("height",20)
		;

		k.append('div')
			.style("width",15)
			.style("height",15)
			.style("background-color",colors(graphnum))
			.style("border","solid 1px")
			.style("float", "left")
		;
		k.append("text")
			.text(" = " + String.fromCharCode(graphnum+65))
			.attr("class", "SubTitle")
			.style("float", "right")
			.style("color",colors(graphnum))
		;
	});



}
function GenerateCategorySelect() {
	//Generates the filter selection window

	//If the filter window is off only show a button
	if (!ConfigChart_Settings[ConfigChart_CurrentIndex].ShowFilters) {
		//Create a button to activate the window
		var sel = d3.select("#tools"+ConfigChart_CurrentIndex)
			.append("button")
				.data([ConfigChart_CurrentIndex])
				.attr("class", "mybutton")
				.style("font-size", 22)
				.style("margin-top", 70)
				.style("border-top-right-radius","0px")
				.style("border-bottom-right-radius","0px")
				.on("click", function (d) {

					ConfigChart_Settings[d].ShowFilters = !ConfigChart_Settings[d].ShowFilters;

					GenerateCharts();
				})
				.text(function(d) {
					//The button's position must be adjusted if the function window is off
					if (ConfigChart_Settings[d].ShowFilters) return "X";
					else return "Filters";
				})
				.style("position", "absolute")

				if (ConfigChart_Settings[ConfigChart_CurrentIndex].ShowFunction)
					sel.style("right", ConfigChart_FunctionWidth + ConfigChart_BorderWidth*2)
				else sel.style("right", 0)
		;
		return;
	}

	var td = d3.select("#tools"+ConfigChart_CurrentIndex).append("td").style("border",0);
	var height = ConfigChart_Height+ConfigChart_AxisWidth + ConfigChart_TitlebarHeight;

	var Table = td.append("table")
		.style("background-color", "#AADDFF")
		.style("height", height)
			
		.style("border-top-left-radius",(ConfigChart_Roundness - ConfigChart_BorderWidth) + "px")
		.style("border-bottom-left-radius",(ConfigChart_Roundness - ConfigChart_BorderWidth) + "px")
	;

	var row0 = Table.append("tr");
	var row1 = Table.append("tr");
	var row2 = Table.append("tr");
	var row3 = Table.append("tr");

	var departments = Array.from(ConfigChart_AllDepartments);
	var catalogs = Array.from(ConfigChart_AllCatalogs);
	var categories = Array.from(ConfigChart_AllCategories);

	departments.sort();
	catalogs.sort();
	categories.sort();

	//Add a null option to the drowpdown menus
	departments.unshift("-select-");
	catalogs.unshift("-select-");
	//categories.unshift("-select-");

	var selectcolor = {"bg":"#AABBFF","fg":"#000000"};
	var deselectcolor = {"bg":"#555555","fg":"#FFFFFF"};

	var tbl, selection;

	//Create a button to toggle the filter window
	selection = row0.append("td")
		.style("vertical-align", "top")
	;
	selection.append("button")
			.data([ConfigChart_CurrentIndex])
			.attr("class", "mybutton")
			.style("margin", 0)
			.on("click", function (d) {

				ConfigChart_Settings[d].ShowFilters = !ConfigChart_Settings[d].ShowFilters;

				GenerateCharts();
			})
			.text(function(d) {
				if (ConfigChart_Settings[d].ShowFilters) return "X";
				else return "Filters";
			})
	;


	//Initialize the random colors
	var colors = d3.scale.category10();
	for (i = 0; i < ConfigChart_Filters[ConfigChart_CurrentIndex].length; ++i) colors(i);

	//Create a filter selection drowpdown menu
	var filters = selection.append('select')
		.data([ConfigChart_CurrentIndex])
		.style("float", "right")
		.style("background-color", function (d) {return colors(ConfigChart_Settings[d].FilterNum);})
		.on("change",function(d) {
			//When the filter is changed don't copy the function over 
			ConfigChart_Functions[d].ignore = true;

			//If we are meant to add a new filter
			if (this.value == 'New') {
				//Duplicate the current filter into the new one
				var lastfilters = ConfigChart_Filters[d][ConfigChart_Settings[d].FilterNum];
				var prevfunctionnum = ConfigChart_Settings[d].FilterNum;
				var prevfunction = ConfigChart_Functions[d][prevfunctionnum].func;
				var curfunctionnum = ConfigChart_Functions[d].length;

				ConfigChart_Filters[d].push({"categories":new Set(),"departments":new Set(),"classes":new Set()});

				var newfilter = ConfigChart_Filters[d][curfunctionnum];

				lastfilters.categories.forEach(function(d) {newfilter.categories.add(d);});
				lastfilters.departments.forEach(function(d) {newfilter.departments.add(d);});
				lastfilters.classes.forEach(function(d) {newfilter.classes.add(d);});

				ConfigChart_Settings[d].FilterNum = curfunctionnum;


				//Copy the current function to the new function

			
				ConfigChart_Functions[d].push({"func":[],"tokens":[],"jump":[],"selectedkey": ConfigChart_Functions[d][prevfunctionnum].selectedkey});
				prevfunction.forEach(function(line) {
					ConfigChart_Functions[d][curfunctionnum].func.push(line);
				});

				ConfigChart_Filters[d][curfunctionnum].func = curfunctionnum;

				Tokenize(d,curfunctionnum);
			}
			//Select the chosen filter
			else ConfigChart_Settings[d].FilterNum = this.value;

			GenerateCharts();
		})
		.attr("name", "filters")
	;

	selection.append('text')
		.style("float", "right")
		.html('Filter:&nbsp')
	;

	selection = filters;

	//Add all of the options to the dropdown
	for (i = 0; i < ConfigChart_Filters[ConfigChart_CurrentIndex].length; ++i) 
		selection.append("option")
			.data([i])
			.attr("value", function(d){return d;})
			.style("background-color", function(d) {return colors(d);})
			.text(String.fromCharCode(i+65))
			.property("selected", function(d) {return d == ConfigChart_Settings[ConfigChart_CurrentIndex].FilterNum;})
	;
	selection.append("option").style("background-color", "#FFFFFF").text("New");	//Add new option

	td = row1.append("td");

	tbl = td
		.append('table')
			.attr("cellpadding", 0)
			.attr("cellspacing", 0)
			.attr("border", 0)
			.style("width", ConfigChart_FunctionWidth)
			.style("height", height*.3)
	;


	//Show/hide the current filter with this button
	var r = tbl.append('tr');
	r.append('td')
		.style("height", "4%")
		.html("Category Filter:<br>")
	;
	r.append('td').append('button')
		.style("float", "right")
		.attr("class", "mybutton")
		.data([{"index":ConfigChart_CurrentIndex}])
		.on('click', function(d) {
			if (ConfigChart_Settings[d.index].HiddenFilters.has(Number(ConfigChart_Settings[d.index].FilterNum))) 
				ConfigChart_Settings[d.index].HiddenFilters.delete(Number(ConfigChart_Settings[d.index].FilterNum));
			else 
				ConfigChart_Settings[d.index].HiddenFilters.add(Number(ConfigChart_Settings[d.index].FilterNum));

			GenerateCharts();
		})
		.text(function(d) {
			if (ConfigChart_Settings[d.index].HiddenFilters.has(Number(ConfigChart_Settings[d.index].FilterNum))) 
				return "show";
			
			return "hide";
			
		});
	;

	//Create category selection
	selection = tbl.append('tr').append('td')	
		.style("height", "92%")
		.attr('colspan',2)
		.append("div")
			.style("overflow-y", "scroll")
			.style("background-color", "#000000")
			.style("padding", "2%")
			.style("font-family", "monospace")
			.style("color", "#FFFFFF")
			.style("vertical-align", "top")
			.style("height", "100%")
	;
	categories.forEach(function (d) {
		selection.append('button')
			.data([{"name":d,"index":ConfigChart_CurrentIndex}])
			.attr("class", "mybutton")
			.style("background-color",function(d) {
				if (ConfigChart_Filters[d.index][ConfigChart_Settings[d.index].FilterNum].categories.has(d.name))
					return selectcolor.bg;
				else return deselectcolor.bg;
			})
			.style("color",function(d) {
				if (ConfigChart_Filters[d.index][ConfigChart_Settings[d.index].FilterNum].categories.has(d.name))
					return selectcolor.fg;
				else return deselectcolor.fg;
			})
			.text(d)
			.on("click",function(d) {
				var categories = ConfigChart_Filters[d.index][ConfigChart_Settings[d.index].FilterNum].categories;
				if (categories.has(d.name))
					categories.delete(d.name);
				else categories.add(d.name);

				GenerateCharts();
			})
		;
	});

	//Create department selection
	td = row2.append("td");
	tbl = td
		.append('table')
			.attr("cellpadding", 0)
			.attr("cellspacing", 0)
			.attr("border", 0)
			.style("width", ConfigChart_FunctionWidth)
			.style("height", height*.3)
	;
	tbl.append('tr').append('td')
		.style("height", "4%")
		.html("Department Filter:<br>");
	selection = tbl.append('tr').append('td')	
		.style("height", "92%")
		.append("div")
			.style("overflow-y", "scroll")
			.style("background-color", "#000000")
			.style("padding", "2%")
			.style("font-family", "monospace")
			.style("color", "#FFFFFF")
			.style("vertical-align", "top")
			.style("height", "100%")
	;
	selection.append('select')
		.data([ConfigChart_CurrentIndex])
		.on("change",function(d) {
			var departments = ConfigChart_Filters[d][ConfigChart_Settings[d].FilterNum].departments;
			if (departments.has(this.value))
				departments.delete(this.value);
			else departments.add(this.value);

			GenerateCharts();
		})
		.attr("name", "Departments")
		.selectAll('option')
		.data(departments)
			.enter()
				.append("option")
					.text(function (d) {
						return d;
					})
	;
	ConfigChart_Filters[ConfigChart_CurrentIndex][ConfigChart_Settings[ConfigChart_CurrentIndex].FilterNum].departments.forEach(function(d) {
		selection.append('button')
		.data([{"name":d,"index":ConfigChart_CurrentIndex}])
			.attr("class", "mybutton")
			.text(d)
			.on("click",function(d) {
				ConfigChart_Filters[d.index][ConfigChart_Settings[d.index].FilterNum].departments.delete(d.name);
				GenerateCharts();
			})
	});

	//Create class selection
	td = row3.append("td");
	tbl = td
		.append('table')
			.attr("cellpadding", 0)
			.attr("cellspacing", 0)
			.attr("border", 0)
			.style("width", ConfigChart_FunctionWidth)
			.style("height", height*.3)
	;
	tbl.append('tr').append('td')
		.style("height", "4%")
		.html("Class Filter:<br>");
	selection = tbl.append('tr').append('td')	
		.style("height", "92%")
		.append("div")
			.attr("id", "ClassSelect"+ConfigChart_CurrentIndex)
			.style("overflow-y", "scroll")
			.style("background-color", "#000000")
			.style("padding", "2%")
			.style("font-family", "monospace")
			.style("color", "#FFFFFF")
			.style("vertical-align", "top")
			.style("height", "100%")
	;
	selection.append('select')
		.data([ConfigChart_CurrentIndex])
		.on("change",function(d) {
			var classes = ConfigChart_Filters[d][ConfigChart_Settings[d].FilterNum].classes;
			if (classes.has(this.value))
				classes.delete(this.value);
			else classes.add(this.value);

			GenerateCharts();
		})
		.attr("name", "Class Numbers")
		.selectAll('option')
		.data(catalogs)
			.enter()
				.append("option")
					.text(function (d) {
						return d;
					})

	;
	ConfigChart_Filters[ConfigChart_CurrentIndex][ConfigChart_Settings[ConfigChart_CurrentIndex].FilterNum].classes.forEach(function(d) {
		selection.append('button')
		.data([{"name":d,"index":ConfigChart_CurrentIndex}])
			.attr("class", "mybutton")
			.text(d)
			.on("click",function(d) {
				ConfigChart_Filters[d.index][ConfigChart_Settings[d.index].FilterNum].classes.delete(d.name);
				GenerateCharts();
			})
	});


}
function GenerateFunctionEdit() {
	//Generates the function window

	//If the function window is disabled
	if (!ConfigChart_Settings[ConfigChart_CurrentIndex].ShowFunction) {
		//Create a button to turn it back on
		var sel = d3.select("#tools"+ConfigChart_CurrentIndex)
			.append("button")
				.data([ConfigChart_CurrentIndex])
				.attr("class", "mybutton")
				.style("font-size", 22)
				.style("margin-top", 10)
				.style("border-top-right-radius","0px")
				.style("border-bottom-right-radius","0px")
				.on("click", function (d) {

					ConfigChart_Settings[d].ShowFunction = !ConfigChart_Settings[d].ShowFunction;

					GenerateCharts();
				})
				.text(function(d) {
					if (ConfigChart_Settings[d].ShowFunction) return "X";
					else return "\u2211 function";
				})
				.style("position", "absolute")

				if (ConfigChart_Settings[ConfigChart_CurrentIndex].ShowFilters)
					sel.style("right", ConfigChart_FunctionWidth + ConfigChart_BorderWidth*2)
				else sel.style("right", 0)
		;
		return;
	}

	var td = d3.select("#tools"+ConfigChart_CurrentIndex).append("td").style("border",0);
	var height = ConfigChart_Height+ConfigChart_AxisWidth + ConfigChart_TitlebarHeight;

	var Table = td.append("table")
		.style("background-color", "#AADDFF")
		.style("height", height)
	;

	//Round the corners if necessary
	if (!ConfigChart_Settings[ConfigChart_CurrentIndex].ShowFilters) {
		Table.style("border-top-left-radius",(ConfigChart_Roundness - ConfigChart_BorderWidth) + "px")
			.style("border-bottom-left-radius",(ConfigChart_Roundness - ConfigChart_BorderWidth) + "px")
		;
	}

	var row1 = Table.append("tr");
	var row2 = Table.append("tr");
	var row3 = Table.append("tr");

	var selectcolor = {"bg":"#AABBFF","fg":"#000000"};
	var deselectcolor = {"bg":"#555555","fg":"#FFFFFF"};

	var tbl, selection;


	var selection = row1.append("td")
		.style("vertical-align", "top")
		.append('div');

	//Button for switching between presets and function view
	selection.append("button")
			.data([ConfigChart_CurrentIndex])
			.attr("class", "mybutton")
			.style("float", "left")
			.on("click", function (d) {

				ConfigChart_Settings[d].ShowFunction = !ConfigChart_Settings[d].ShowFunction;

				GenerateCharts();
			})
			.text(function(d) {
				if (ConfigChart_Settings[d].ShowFunction) return "X";
				else return "\u2211 function";
			})
	;

	//Button for toggling between presets and function edit
	row1.append("td")
		.style("vertical-align", "top")
		.append("button")
			.data([ConfigChart_CurrentIndex])
			.attr("class", "mybutton")
			.style("float", "right")
			.on("click", function (d) {

				ConfigChart_Settings[d].ShowPresets = !ConfigChart_Settings[d].ShowPresets;

				GenerateCharts();
			})
			.text(function(d) {
				if (ConfigChart_Settings[d].ShowPresets) return "\u2211 function";
				else return "Presets";
			})
	;

	//If the presets window is active, generate it and exit
	if (ConfigChart_Settings[ConfigChart_CurrentIndex].ShowPresets) {
		GeneratePresets(row2,height);
		return;
	}


	var func = "";
	var funcnum = ConfigChart_Settings[ConfigChart_CurrentIndex].FilterNum;

	//Convert the functions tokens to plain html
	ConfigChart_Functions[ConfigChart_CurrentIndex][funcnum].tokens.forEach(function (d) {
		for (i = 0; i < d.length; ++i) {
			func += d[i] + " ";
		}
		func += "<br>";
	});


	var td1 = row2.append("td").attr('colspan',2);
	var functable = td1
		.append('table')
			.attr("cellpadding", 0)
			.attr("cellspacing", 0)
			.attr("border", 0)
			.style("width", ConfigChart_FunctionWidth)
			.style("height", height*.7)
	;

	//Create the editable function div
	functable.append('tr').append('td')
		.html("Summation Function:<br>");
	functable.append('tr').append('td')	
		.style("height", "92%")
		.append("div")
			.attr("id", "EditFunction"+ConfigChart_CurrentIndex)
			.style("overflow-y", "scroll")
			.style("background-color", "#000000")
			.style("padding", "2%")
			.attr("contentEditable", "true")
			.style("font-family", "monospace")
			.style("color", "#FFFFFF")
			.style("vertical-align", "top")
			.style("height", "100%")
			.html(func)
			.on("keyup",keypressed)
	;

	//For some reason double clicking wants to select a trailing space on some browsers. This removes that space.
	document.getElementById("EditFunction"+ConfigChart_CurrentIndex).ondblclick = function() {
		if (document.selection && document.selection.empty) {
			document.selection.empty();
		}
		else if (window.getSelection) {
			var s = window.getSelection();
			var range = s.getRangeAt(0);
			if (typeof range.endContainer.wholeText === 'undefined') return;
			if (range.endContainer.wholeText[0] == ' ' || range.endContainer.wholeText[0] == String.fromCharCode(160)) {
				range.setEnd(range.endContainer, range.endOffset-1);
				s.removeAllRanges();
				s.addRange(range);
			}
		}
	}

	//Button to run the function manually
	functable.append('tr').append("td")
		.style("height", "4%")
		.append("button")
		.data([ConfigChart_CurrentIndex])
		.attr("class", "mybutton")
		.style("margin-top", 2)
		.on("click", function (d) {
			var td = d3.select("#EditFunction"+d)[0][0];
			var funcnum = ConfigChart_Filters[d][ConfigChart_Settings[d].FilterNum].func;

			//Get the most current function
			ConfigChart_Functions[d][funcnum].func = td.innerText.split("\n");

			Tokenize(d,funcnum);

			GenerateCharts();
		})
		.html("Run")
	;

	//Generate the error console
	var consoletable = row3.append("td").attr('colspan',2).append('table')
		.attr("cellpadding", 0)
		.attr("cellspacing", 0)
		.attr("border", 0)
		.style("width", ConfigChart_FunctionWidth)
		.style("height", height*.2)
	;

	consoletable.append('tr').append('td')
		.style("height", "4%")
		.html("Error Console:<br>")
	;
	consoletable.append('tr').append('td')
	.style("height", "96%")
		.append("div")
			.attr("id", "ErrorConsole"+ConfigChart_CurrentIndex)
			.style("overflow-y", "scroll")
			.style("background-color", "#000000")
			.style("padding", "2%")
			.style("font-family", "monospace")
			.style("color", "#FFFFFF")
			.style("vertical-align", "top")
			.style("height", "100%")
			.html("")

	;


}

function ParseData() {
	//Evaluates the data for all functions

	if (ConfigChart_json.data.length == 0) return [];

	//Initialze the results
	var results = [];
	results.maxvalue = -Infinity;
	results.minvalue = Infinity;

	//Keep track of what term numbers are needed
	var termsSet = new Set();

	//For each filter evaluate the data
	for (filternum = 0; filternum < ConfigChart_Filters[ConfigChart_CurrentIndex].length;++filternum) {

		//clear the symbol table for a new computation
		ConfigChart_SymbolTable = [];		

		//Initialize the results for a sigle filter
		var data = [];
		data.maxvalue = -Infinity;
		data.minvalue = Infinity;

		var CurrentTerm = null;

		var func = ConfigChart_Functions[ConfigChart_CurrentIndex][filternum];

		//For each datapoint run the custom function
		ConfigChart_json.data.forEach(function (d) {
			//If this datapoint does not pass the filters skip it
			if (!PassesFilter(d,filternum)) return;

			//foe each new term start a new sum
			if (d.Term != CurrentTerm) {
				if (ConfigChart_SymbolTable["prev"] > data.maxvalue) data.maxvalue = ConfigChart_SymbolTable["prev"];
				if (ConfigChart_SymbolTable["prev"] < data.minvalue) data.minvalue = ConfigChart_SymbolTable["prev"];

				//clear the symbol table for a new computation
				ConfigChart_SymbolTable = [];
				CurrentTerm = d.Term;

				termsSet.add(CurrentTerm);

				data.push({"value":0, "term" : CurrentTerm, "components": []});
			}

			//compute the value for this datapoint
			var newVal = ComputeFunction(d,func);

			//If there was no change consider the datapoint invalid
			if (newVal != data[data.length-1].value) {
				data[data.length-1].components.push({"key":d.Department + " " + d.Catalog,"value": (newVal - data[data.length-1].value)});
				data[data.length-1].value = ConfigChart_SymbolTable["prev"];
			}

		});

		//Update the max and min computed values
		if (ConfigChart_SymbolTable["prev"] > data.maxvalue) data.maxvalue = ConfigChart_SymbolTable["prev"];
		if (ConfigChart_SymbolTable["prev"] < data.minvalue) data.minvalue = ConfigChart_SymbolTable["prev"];


		if (data.length == 1) {
			data.maxvalue = data[0].value+1;
			data.minvalue = data[0].value-1;
		}

		//Update the final max and min computed values
		if (results.maxvalue < data.maxvalue) results.maxvalue = data.maxvalue;
		if (results.minvalue > data.minvalue) results.minvalue = data.minvalue;

		results.push(data);
	}
	results.terms = Array.from(termsSet);
	results.terms.sort();
	return results;

}
function ComputeFunction (data,func) {
	//Compute the a value with the function 'func' using the data in 'data'

	var NumIterations = 0;
	var values = data.Values;
	var categories = ConfigChart_json.classes[data.Department + " " + data.Catalog]

	for (i = 0; i < values.length; ++i) 
		ConfigChart_SymbolTable[values[i].key] = Number(values[i].value);
	
	for (i = 0; i < func.tokens.length; ++i) {

		var d = func.tokens[i];

		if (typeof d === 'undefined' || d.length == 0) continue; //Skip empty lines

		//Infinite loop protection. I disabled it to allow for more exotic programs. Reenable it if you so desire
		/*
		++NumIterations;
		if (NumIterations > func.tokens.length*50) {	//If we iterate 50 times more than the length of the function stop
			alert("Limit Reached. You might have created an infinite loop");
			return ConfigChart_SymbolTable["prev"];
		}
		*/

		//Create symbols for everything we see
		d.forEach(function (op) {
			if (typeof ConfigChart_SymbolTable[op] === 'undefined') {
				if (!isNaN(op)) ConfigChart_SymbolTable[op] = Number(op);
				else ConfigChart_SymbolTable[op] = 0;
			}
		});

		//If this is a label
		if (d.length == 1) continue;

		//If this is a return statement
		if (d[0] == "return") {
			ConfigChart_SymbolTable["prev"] = ConfigChart_SymbolTable[d[1]];
			return ConfigChart_SymbolTable["prev"];
		}

		//If this is a branch
		if (d[0] == "if" || d[0] == "bif") {

			var IsTrue = false;
			var jumptarget = i;

			if (typeof func.jump[d[4]] != 'undefined')
				jumptarget = func.jump[d[4]];
			else AddError("unknown label '" + d[4] + "'",i);

			if (d[1] == "category") {
				if (d[2] == "=" || d[2] == "==")
					IsTrue = categories.indexOf(d[3]) != -1;
				else if (d[2] == "!=")
					IsTrue = categories.indexOf(d[3]) == -1;
				else AddError("cannot perform operation '" + d[2] + "' on type 'category'",i)

			}
			else if (d[3] == "category") {
				if (d[2] == "=" || d[2] == "==")
					IsTrue = categories.indexOf(d[1]) != -1;
				else if (d[2] == "!=")
					IsTrue = categories.indexOf(d[1]) == -1;
				else AddError("cannot perform operation '" + d[2] + "' on type 'category'",i)
				
			}

			else if (d[1] == "department") {
				if (d[2] == "=" || d[2] == "==")
					IsTrue = data.Department == d[3];
				else if (d[2] == "!=")
					IsTrue = data.Department != d[3];
				else AddError("cannot perform operation '" + d[2] + "' on type 'category'",i)

			}
			else if (d[3] == "department") {
				if (d[2] == "=" || d[2] == "==")
					IsTrue = data.Department == d[1];
				else if (d[2] == "!=")
					IsTrue = data.Department != d[1];
				else AddError("cannot perform operation '" + d[2] + "' on type 'category'",i)
				
			}

			else if (d[1] == "catalog") {
				if (d[2] == "=" || d[2] == "==")
					IsTrue = data.Catalog == d[3];
				else if (d[2] == "!=")
					IsTrue = data.Catalog != d[3];
				else AddError("cannot perform operation '" + d[2] + "' on type 'category'",i)

			}
			else if (d[3] == "catalog") {
				if (d[2] == "=" || d[2] == "==")
					IsTrue = data.Catalog == d[1];
				else if (d[2] == "!=")
					IsTrue = data.Catalog != d[1];
				else AddError("cannot perform operation '" + d[2] + "' on type 'category'",i)
				
			}

			else {
				if (d[2] == "=" || d[2] == "==")
					IsTrue = ConfigChart_SymbolTable[d[1]] == ConfigChart_SymbolTable[d[3]];
				else if (d[2] == "!=")
					IsTrue = ConfigChart_SymbolTable[d[1]] != ConfigChart_SymbolTable[d[3]];
				else if (d[2] == "<")
					IsTrue = ConfigChart_SymbolTable[d[1]] < ConfigChart_SymbolTable[d[3]];
				else if (d[2] == ">")
					IsTrue = ConfigChart_SymbolTable[d[1]] > ConfigChart_SymbolTable[d[3]];
				else if (d[2] == "<=")
					IsTrue = ConfigChart_SymbolTable[d[1]] <= ConfigChart_SymbolTable[d[3]];
				else if (d[2] == ">=")
					IsTrue = ConfigChart_SymbolTable[d[1]] >= ConfigChart_SymbolTable[d[3]];
				else if (d[2] == "|" || d[2] == "||")
					IsTrue = ConfigChart_SymbolTable[d[1]] || ConfigChart_SymbolTable[d[3]];
				else if (d[2] == "&" || d[2] == "&&")
					IsTrue = ConfigChart_SymbolTable[d[1]] && ConfigChart_SymbolTable[d[3]];
				else AddError("unknown comparator '" + d[2] + "'",i)
			}

			if (IsTrue) i = jumptarget;
			
			continue;
		}

		//If this is a jump to a label
		if (d[0] == "jump") {
			if (typeof func.jump[d[1]] != 'undefined')
				i = func.jump[d[1]];
			else AddError("unknown label '" + d[1] + "'",i);
			continue;
		}

		//If this is a type of assignment
		if (d[1] == "=") {
			//If this is a simple assignment
			if (typeof d[3] === 'undefined') {
				ConfigChart_SymbolTable[d[0]] = ConfigChart_SymbolTable[d[2]];
				continue
			}
			if (d[2] === 'sqrt') {
				if (ConfigChart_SymbolTable[d[3]] < 0) AddError("cannot take the square root of " + ConfigChart_SymbolTable[d[3]],i);
				ConfigChart_SymbolTable[d[0]] = Math.sqrt(ConfigChart_SymbolTable[d[3]]);
				continue;
			}

			if (typeof ConfigChart_SymbolTable[d[4]] === 'undefined') {
				AddError("missing an operand",i);
				continue
			}

			var tmp = ConfigChart_SymbolTable[d[2]];

			//If the operands are categories
			if (d[2] == "category") {
				if (d[3] == "=" || d[3] == "==")
					tmp = categories.indexOf(d[4]) != -1;
				else if (d[3] == "!=")
					tmp = categories.indexOf(d[4]) == -1;
				else AddError("cannot perform operation '" + d[3] + "' on type 'category'",i)

			}
			else if (d[4] == "category") {
				if (d[3] == "=" || d[3] == "==")
					tmp = categories.indexOf(d[2]) != -1;
				else if (d[3] == "!=")
					tmp = categories.indexOf(d[2]) == -1;
				else AddError("cannot perform operation '" + d[3] + "' on type 'category'",i)

			}
			else {

				//Else perform a binary operation and save the result
				 
				if (d[3] == "+")
					tmp += ConfigChart_SymbolTable[d[4]];
				else if (d[3] == "-")
					tmp -= ConfigChart_SymbolTable[d[4]];
				else if (d[3] == "*")
					tmp *= ConfigChart_SymbolTable[d[4]];
				else if (d[3] == "/")
					tmp /= ConfigChart_SymbolTable[d[4]];
				else if (d[3] == "%")
					tmp %= ConfigChart_SymbolTable[d[4]];
				else if (d[3] == "==" || d[3] == "=")
					tmp = tmp == ConfigChart_SymbolTable[d[4]];
				else if (d[3] == "!=")
					tmp = tmp != ConfigChart_SymbolTable[d[4]];
				else if (d[3] == "<")
					tmp = tmp < ConfigChart_SymbolTable[d[4]];
				else if (d[3] == ">")
					tmp = tmp > ConfigChart_SymbolTable[d[4]];
				else if (d[3] == "<=")
					tmp = tmp <= ConfigChart_SymbolTable[d[4]];
				else if (d[3] == ">=")
					tmp = tmp >= ConfigChart_SymbolTable[d[4]];
				else if (d[3] == "|" || d[3] == "||")
					tmp = tmp | ConfigChart_SymbolTable[d[4]];
				else if (d[3] == "&" || d[3] == "&&")
					tmp = tmp & ConfigChart_SymbolTable[d[4]];
				else AddError("unknown binary operator '" + d[3] + "'",i);
			}

			ConfigChart_SymbolTable[d[0]] = tmp;

			continue;
		}
		AddError("syntax error"	,i);
	}

	//Reset the iteration-specific values so they don't carry over
	for (i = 0; i < values.length; ++i) 
		ConfigChart_SymbolTable[values[i].key] = 0;

	return ConfigChart_SymbolTable["prev"];
}
function Tokenize(index,funcnum) {
	//Divides the current function string array into tokens
	//Also creates a jumptable for the function

	var data = ConfigChart_Functions[index][funcnum];

	data.tokens = [];

	var Function = data.func;

	//Loop though each line
	Function.forEach(function(d,i) {

		if (i == Function.length-1 && d.length == 0) return;
		if (d.length == 0) data.tokens[i] = [""];

		d = d.toLowerCase();	//Don't allow uppercase for the sake of simplicity

		//Split each line into tokens
		data.tokens[i] = d.match(/[a-zA-Z]+|[0-9]+|(<=|>=|!=|==|\|\||&&)|(=|\+|\||-|\*|\/|<|>|%|&|\|)/g);

		if (data.tokens[i] == null) {
			data.tokens[i] = [""];
			if (d.length > 0)
				console.log("unexpected token at line " + i);
		}
	});

	//Generate a jump table
	data.jump = [];
	data.tokens.forEach(function(d,i) {
		if (d.length == 0) return;

		data.jump[d[0]] = i;
	});
}

function keypressed() {
	//When a key is pressed start a timer that waits and then applies syntax highlighting
	if (ConfigChart_KeyPressEvent != null)
		window.clearTimeout(ConfigChart_KeyPressEvent);

	ConfigChart_KeyPressEvent = setTimeout(SyntaxHighlighting, 5000);
}
function SyntaxHighlighting() {
	//Apply syntax highlighting to all functions

	var scrollpos = window.pageYOffset;

	//For each chart that exists
	for (index = 0; index < ConfigChart_NumCharts; index++ ) {
		var td = d3.select("#EditFunction"+index)[0][0];	//Get the function div associated with the chart

		if (td == null) continue;	//If the div does not exist continue on

		var funcnum = ConfigChart_Settings[index].FilterNum;

		//Save the function
		ConfigChart_Functions[index][funcnum].func = td.innerText.split("\n");

		//Tokenize the function to clean it and remove bad tokens
		Tokenize(index,funcnum);

		//Color each token based on it's line position and string
		var func = "";
		ConfigChart_Functions[index][funcnum].tokens.forEach(function (d) {
			for (i = 0; i < d.length; ++i) {
				if (d[i] == "if" || d[i] == "bif" || d[i] == "return" || d[i] == "jump" || d[i] == "sqrt") 
					func += '<span style="color:lightgreen">' + d[i] + '</span> ';

				else if (d[i] == "prev" || ConfigChart_AllValueKeys.has(d[i])) 
					func += '<span style="color:cyan">' + d[i] + '</span> ';

				else if (d.length > 4 && (i == 1 || i == 3) && (
						(d[1] == "category" || d[3] == "category") 
						|| (d[1] == "department" || d[3] == "department")
						|| (d[1] == "catalog" || d[3] == "catalog")
					))
					func += '<span style="color:orange">' + d[i] + '</span> ';

				else if (d.length > 4 && (i == 2 || i == 4) && (d[2] == "category" || d[4] == "category"))
					func += '<span style="color:orange">' + d[i] + '</span> ';

				else if (d.length == 1 || (i == 4 && (d[0] == "if" || d[0] == "bif")) || (i == 1 && d[0] == "jump")) 
					func += '<span style="color:yellow">' + d[i] + '</span> ';
								
				else func += d[i] + " ";

			}
			func += "<br>";
		});

		//Apply the changes to the div
		d3.select("#EditFunction"+index).html(func);
	}

	//This deselects everything. I tried making the cursor caret stay in the correct place after syntax hightlighting, but trust me that's not happening. 
	BlurAll();	

	window.scrollTo(0,scrollpos);
}

function AddError (error, index) {
	//Add the error message 'error' to the error console for chart 'index'

	if (ConfigChart_Settings[ConfigChart_CurrentIndex].ShowPresets) return;

	d3.select('#ErrorConsole'+ConfigChart_CurrentIndex)[0][0].innerHTML += error + " at line " + (index+1) + "<br>";

	var filternum = ConfigChart_Settings[ConfigChart_CurrentIndex].FilterNum;
	var funcnum = ConfigChart_Filters[ConfigChart_CurrentIndex][filternum].func;

	//Clear the function at this line since it is faulty
	ConfigChart_Functions[ConfigChart_CurrentIndex][funcnum].tokens[index] = [""];
}
function BlurAll() {
	//Clear all selections
	var tmp = document.createElement("input");
	document.body.appendChild(tmp);
	tmp.focus();
	document.body.removeChild(tmp);
}
function ApplyPreset(d,func) {
	//Applies the function 'func' as a preset
	var curfunc = ConfigChart_Functions[d][ConfigChart_Settings[d].FilterNum];
	curfunc.func = func;
	Tokenize(d,ConfigChart_Settings[d].FilterNum);
	GenerateCharts();
}
function AddChart() {

	//Adds a new chart by duplicating the previous one if one exists.

	ConfigChart_Settings[ConfigChart_NumCharts] = {"ShowFilters":false,"ShowFunction":false, "ShowPresets":true, "FilterNum":0, "HiddenFilters": new Set()};
	ConfigChart_Filters[ConfigChart_NumCharts] = [];


	//If there at least one chart in existance
	if (ConfigChart_NumCharts > 0) {
		//Copy over all settings
		ConfigChart_Settings[ConfigChart_NumCharts].ShowFilters = ConfigChart_Settings[ConfigChart_NumCharts-1].ShowFilters;
		ConfigChart_Settings[ConfigChart_NumCharts].ShowFunction = ConfigChart_Settings[ConfigChart_NumCharts-1].ShowFunction;
		ConfigChart_Settings[ConfigChart_NumCharts].ShowPresets = ConfigChart_Settings[ConfigChart_NumCharts-1].ShowPresets;
		ConfigChart_Settings[ConfigChart_NumCharts].FilterNum = ConfigChart_Settings[ConfigChart_NumCharts-1].FilterNum;
		ConfigChart_Settings[ConfigChart_NumCharts-1].HiddenFilters.forEach(function(d) {
			ConfigChart_Settings[ConfigChart_NumCharts].HiddenFilters.add(d);
		});

		//Copy over every function
		ConfigChart_Functions[ConfigChart_NumCharts] = [];

		ConfigChart_Functions[ConfigChart_NumCharts-1].forEach(function (d,i) {
			ConfigChart_Functions[ConfigChart_NumCharts].push({"func":[],"tokens":[],"jump":[],"ignore":true,"selectedkey":d.selectedkey});
			d.func.forEach(function(d) {
				ConfigChart_Functions[ConfigChart_NumCharts][i].func.push(d);
			});
			Tokenize(ConfigChart_NumCharts,i);
		});

		//Copy over all filters
		var oldfilter = ConfigChart_Filters[ConfigChart_NumCharts-1];
		var newfilter = ConfigChart_Filters[ConfigChart_NumCharts];
		for (i = 0; i < oldfilter.length; ++i) {
			newfilter.push({"categories":new Set(),"departments":new Set(),"classes":new Set()});

			oldfilter[i].categories.forEach(function(d) {newfilter[i].categories.add(d);});
			oldfilter[i].departments.forEach(function(d) {newfilter[i].departments.add(d);});
			oldfilter[i].classes.forEach(function(d) {newfilter[i].classes.add(d);});

		}
	}
	else {
		ConfigChart_Filters[ConfigChart_NumCharts].push({"categories":new Set(),"departments":new Set(),"classes":new Set()});
	}

	++ConfigChart_NumCharts;
	GenerateCharts();
}
function RemoveChart(index) {
	//Removes a chart
	ConfigChart_Functions.splice(index,1);
	ConfigChart_Settings.splice(index,1);
	ConfigChart_Filters.splice(index,1);


	--ConfigChart_NumCharts;
	GenerateCharts();
}
function PassesFilter(d,filternum) {
	//Returns true if the datapoint d passes the filter 'filternum'

	//Get the filter to use
	var filters = ConfigChart_Filters[ConfigChart_CurrentIndex][filternum];

	passed = false;

	//d must pass all three tests to suceed
	for (j = 0; j < ConfigChart_json.classes[d.Department + " " + d.Catalog].length; ++j) 
		if (filters.categories.has(ConfigChart_json.classes[d.Department + " " + d.Catalog][j])) passed = true;

	if (filters.categories.size != 0 && !passed) return false;

	if (filters.departments.size != 0 && !filters.departments.has(d.Department)) return false;
	if (filters.classes.size != 0 && !filters.classes.has(d.Catalog)) return false;

	return true;
}
function GeneratePresets(row1, height) {
	//A helper function to generate the buttons for the function presets window
	var td1 = row1.append("td").attr('colspan',2);
	var tbl = td1
	.append('table')
		
		.style("width", ConfigChart_FunctionWidth)
		.style("height", height*.7)
	;
	var sel = tbl.append('tr').append('td');
	sel.append('text')
		.attr("class", "SubTitle")
		.attr('type', "text")
		.style("width", "20%")
		.text("Value:")
	;

	//Create a dropdown to select the name of the key to apply the function on
	var curfunc = ConfigChart_Functions[ConfigChart_CurrentIndex][ConfigChart_Settings[ConfigChart_CurrentIndex].FilterNum];
	var keys = sel.append('select')
		.data([ConfigChart_CurrentIndex])
		.style("float", "right")
		.attr('id', "PresetValue"+ConfigChart_CurrentIndex)
		.on("change",function(d) {
			this.value;

			var curfunc = ConfigChart_Functions[d][ConfigChart_Filters[d][ConfigChart_Settings[d].FilterNum].func];

			curfunc.selectedkey = this.value;
		})
		.attr("name", "keys")
	;
	//Add every possible value key as an option
	ConfigChart_AllValueKeys.forEach(function(d) {
		keys.append("option")
			.data([{"me": d, "sel":curfunc.selectedkey}])
			.attr("value", d)
			.text(d)
			.property("selected", function(d) {return d.me == d.sel})
		;
	});

	//The following are buttons for generating the predefined functions

	tbl.append('tr').append('td').append('button')
		.attr("class", "mybutton")
		.data([ConfigChart_CurrentIndex])
		.style("width", "100%")
		.text("Basic Summation")
		.on('click', function (d) {
			var e = document.getElementById('PresetValue'+d); var val = e.options[e.selectedIndex].value;
			var func = [
				("prev = prev + " + val),
				"return prev"
			];
			ApplyPreset(d,func);
		})
	;
	tbl.append('tr').append('td').append('button')
		.attr("class", "mybutton")
		.data([ConfigChart_CurrentIndex])
		.style("width", "100%")
		.text("Count")
		.on('click', function (d) {
			var e = document.getElementById('PresetValue'+d); var val = e.options[e.selectedIndex].value;
			var func = [
				"n = n + 1",
				"return n"
			];
			ApplyPreset(d,func);
		})
	;
	tbl.append('tr').append('td').append('button')
		.attr("class", "mybutton")
		.data([ConfigChart_CurrentIndex])
		.style("width", "100%")
		.text("Average")
		.on('click', function (d) {
			var e = document.getElementById('PresetValue'+d); var val = e.options[e.selectedIndex].value;
			var func = [
				"n = n + 1",
				"tmp = prev / n",
				"prev = prev - tmp", 
				"tmp = " + val + " / n",
				"prev = prev + tmp",
				"return prev"
			];
			ApplyPreset(d,func);
		})
	;
	tbl.append('tr').append('td').append('button')
		.attr("class", "mybutton")
		.data([ConfigChart_CurrentIndex])
		.style("width", "100%")
		.text("Local Max")
		.on('click', function (d) {
			var e = document.getElementById('PresetValue'+d); var val = e.options[e.selectedIndex].value;
			var func = [
				"n = n + 1",
				"if n = 1 init", 
				"if max > " + val + " done", 
				"max = " + val,"done", 
				"return max", "init", 
				"max = " + val, 
				"jump done"
			];
			ApplyPreset(d,func);
		})
	;
	tbl.append('tr').append('td').append('button')
		.attr("class", "mybutton")
		.data([ConfigChart_CurrentIndex])
		.style("width", "100%")
		.text("Local Min")
		.on('click', function (d) {
			var e = document.getElementById('PresetValue'+d); var val = e.options[e.selectedIndex].value;
			var func = [
				"n = n + 1",
				"if n = 1 init", 
				"if min < " + val + " done",
				"min = " + val,
				"done", 
				"return min",
				"init", "min = " + val, 
				"jump done"
			];
			ApplyPreset(d,func);
		})
	;

	//DOES NOT WORK! Not even sure if it is even possible to do standard deviation as a summation
	tbl.append('tr').append('td').append('button')
		.attr("class", "mybutton")
		.data([ConfigChart_CurrentIndex])
		.style("width", "100%")
		.text("Standard Deviation")
		.on('click', function (d) {
			var e = document.getElementById('PresetValue'+d); var val = e.options[e.selectedIndex].value;
			var func = [
				"n = n + 1",
				"sum = sum + " + val,
				"sqr = " + val + " * " + val,
				"sqrsum = sqr + sqrsum",

				"tmpa = n * srqsum",
				"tmpb = sum * sum",

				"tmpa = tmpb - tmpa",

				"tmpa = sqrt tmpa",
				"tmpa = tmpa / n",
				"return tmpa"
			];
			ApplyPreset(d,func);
		})
	;


}