/*
Author: Dominic Nicholson
Description:
This program is designed to display a graph that shows the flow of students from course to
course each semester within a department. 
It reads enrollment data from all students in the 'Sankey_FilterBy' deparment, and then
proceeeds to aggregate the data into classes. For each class a rectangle is drawn whose
height represents the relative size of the class. Then for each student enrolled in that course, 
this program will check previous semesters for prerequisite class enrollment.
(if 'Sankey_OnlyPrereqLinks' is false all previously enrolled classes will be checked) 
For each prerequisite class a link will be created or added to, showing the flow of classes into, and out of, each course.
*/

var Sankey_hostname = "http://dominicnicholson.no-ip.biz/RegVis/";
queue()
	.defer(d3.csv, Sankey_hostname + "Reg.csv")
	.defer(d3.csv, Sankey_hostname + "PreReq.csv")
	.defer(d3.csv, Sankey_hostname + "Courses.csv")
	.defer(d3.csv, Sankey_hostname + "Students.csv")
	.defer(d3.csv, Sankey_hostname + "Graduates.csv")
	.await(InitSankey);


var Sankey_Width,Sankey_Height;					//Read-only. These variables will store the final width/height of the whole diagram
var Sankey_WPadding = 150;						//The width of each node element
var Sankey_HPadding = 50;						//The height of each node element
var Sankey_NodeHeight = .9;						//a value [0-1] indicating how much of the 'Sankey_HPadding' space a node can use
var Sankey_NodeWidth = .2;						//a value [0-1] indicating how much of the 'Sankey_WPadding' space a node can use
var Sankey_GridOpacity = .025;					//The opacity of the background grid
var Sankey_MaxPastChecking = 20;				//How many past semesters should we check for prerequisite links? 
var Sankey_LHS = "catalog";						//The left hand side category
var Sankey_Cont;								//A container to hold the diagram
var Sankey_TooltipInitialPos = {"x":0,"y":0};	//The initial coordinates of the info window
var Sankey_TooltipStartDrag = {"x":0,"y":0};	//For making the info window draggable
var Sankey_SelectedStudents = new Set();		//Contains a list of SIDs to filter by
var Sankey_Enrollment;							//Sankey_Enrollment[SID][TERM] = A set of classes student 'SID' took during term 'TERM'
var Sankey_FilterBy = "CS";						//The subject to filter by
var Sankey_ShowNumbers = false;					//Whether or not to display the enrollment totals for each class
var Sankey_OnlyPrereqLinks = true;				//Only create links between classes 'A' and 'B' if 'A' is a prerequisite for 'B' where 'A' is taken in an earlier term than 'B' 
var Sankey_OnlyPrereqClasses = true;			//Only generate nodes for the diagram if the class they represent is a prerequisite, or requires prerequisite classes.
var Sankey_ShowRetakes = true;					//Show links between the same class?
var Sankey_IsSerial = true;						//Only link classes to the nearest semester that the student took a valid class
var Sankey_SidebarToggle = false;				//Toggles the sidebar 
var Sankey_FilterToggle = true;					//Toggles whether the diagram should be filtered based on 'Sankey_SelectedStudents'
var Sankey_LastSearchString = "";				//The last search string used
var Sankey_CurrentTooltipData;					//The tooltip data currently in use

//These variables are so we don't have to pass their references to every function that needs 
//the unfiltered data
var Sankey_rawReg;
var Sankey_rawPrereqs;
var Sankey_rawClasses;
var Sankey_rawStudents;
var Sankey_rawSIDs;

//Parse the data for use in the diagram
function InitSankey(error, Data, prereqData, classes, studentsData,graduatesData,Subject) {
	if (error) throw error;
	if (typeof Subject != 'undefined') Sankey_FilterBy = Subject;

	var prereq = FilterPrereqs(prereqData);

	graduates = [];
	graduatesData.forEach(
		function (d) {
			graduates.push(d["student_id"]);
		}
	);

	SIDs = [];					//for gatering all unique student ID's
	students = [];				//for finding student-specific data

	studentsData.forEach(
		function (d) {
			SIDs.push(d["student_id"]);
			students[d["student_id"]] = {
				"name":(d["student_fName"] + " " + d["student_lName"]),
				"gender":d["gender"],
				"ethnicity":d["ethnicity"],
				"finalterm":null
			};
			if (graduates.indexOf(d["student_id"]) != -1)
				students[d["student_id"]].graduate = true;
			else students[d["student_id"]].graduate = false;

			students[d["student_id"]].finalterm = null;
		}
	);

	//Filter out duplicate classes
	var last = "";
	classes = classes.filter(
		function (d) {
			
			if (d.subject == Sankey_FilterBy && d.catalog != last) {
				last = d.catalog;
				return d;
			}
			return;
		}
	);


	Data.forEach(
		function (d) {
			//Save the most recent term that this student was enrolled in
			var student = students[d["student_id"]];

			if (student.finalterm == null)
				student.finalterm = d.term;
			else {
				if (student.finalterm < d.term)
					student.finalterm = d.term;
			}
		}
	);


	Sankey_rawReg = Data;
	Sankey_rawPrereqs = prereq;
	Sankey_rawClasses = classes;
	Sankey_rawStudents = students;
	Sankey_rawSIDs = SIDs;
	Sankey_rawGraduates = graduates;

	//Generate the d3js to display the data
	GenerateSankey();

}

function GenerateSankey() {
	//Generates a sankey diagram based on Sankey_ClassTotals, terms, and classes


	var Data = Sankey_rawReg;
	var prereq = Sankey_rawPrereqs;
	var classData = Sankey_rawClasses;
	var students = Sankey_rawStudents;
	var SIDs = Sankey_rawSIDs;

	Sankey_Enrollment = [];

	var termsSet = new Set();		//Set for gatering all unique term numbers

	//Only show classes for a single subject
	var Filtered = Data.filter (
		function (d) {
			//Filter out non-prereq classes if applicable
			if ((!Sankey_OnlyPrereqClasses || prereq["nodes"].indexOf(d["subject"]+d["catalog"]) != -1)) {
				return d;
			}
			return;
		}
	);

	var classes = [];
	classData.forEach(
		function (d) {
			if (Sankey_OnlyPrereqClasses) {
				if (prereq["nodes"].indexOf(d["subject"]+d["catalog"]) != -1) classes.push(d.catalog);
			}
			else classes.push(d.catalog);
		}
	);

	var LastStudent = -1;	//Keep track of when a unique SID appears

	//Store all unique classes, terms, SID's, and construct the 3D array
	//Sankey_Enrollment[SID][term] = a list of enrolled classes for student 'SID' in term 'term'
	Filtered.forEach(
		function (d) {
			if (Sankey_SelectedStudents.size == 0 || !Sankey_FilterToggle || Sankey_SelectedStudents.has(d["student_id"])) {
				if (d["student_id"] != LastStudent) {
					//This is a unique student
					Sankey_Enrollment[d["student_id"]] = [];

					LastStudent = d["student_id"];
				}
				if (typeof Sankey_Enrollment[d["student_id"]][d["term"]] === 'undefined') {
					Sankey_Enrollment[d["student_id"]][d["term"]] = [];
				}

				var courseexists = false;
				Sankey_Enrollment[d["student_id"]][d["term"]].forEach(function(t) {
					if (d["catalog"] == t["catalog"]) courseexists = true;
				});


				//If the grade is NULL, this may be a duplicate. Save it only if it is unique
				if (d["grade"] == "NULL") {
					
					if (!courseexists) {
						Sankey_Enrollment[d["student_id"]][d["term"]].push({
							"catalog":d[Sankey_LHS],
							"grade":d["grade"]
						});
					}
				}
				
				else {
					//If the course exists already, overwrite it with a valid grade
					if (courseexists) {
						Sankey_Enrollment[d["student_id"]][d["term"]].forEach(function(t) {
							if (d["catalog"] == t["catalog"]) Sankey_Enrollment[d["student_id"]][d["term"]].grade = d["grade"];
						});
					}
					else {
						Sankey_Enrollment[d["student_id"]][d["term"]].push({
							"catalog":d[Sankey_LHS],
							"grade":d["grade"]
						});
					}
				}
				termsSet.add(d["term"]);
			}
		}
	);


	//Convert the sets to arrays for indexing
	var terms = Array.from(termsSet);

	//Sort the terms and classes for display
	terms.sort();

	var Sankey_ClassTotals = [];	//2D array to hold student data aggregated by class

	//Initialize the 3D array so [classname][term].camefrom = an array of class names that preceded the taking of 'classname' in 'term'
	for (j = 0; j < classes.length; ++j) {
		Sankey_ClassTotals[classes[j]] = [];
		for (i = 0; i < terms.length; ++i)
			Sankey_ClassTotals[classes[j]][terms[i]] = {"sum" : 0, "camefrom" : [], "grades": [], "in" : new Set(), "out" : new Set(),"dropouts" : new Set()};
	}

	//Take the enrollment data and generate class totals as well as 'camefrom' data
	//This is the bulk of the data processing work
	for (t = 0; t < terms.length; ++t) {
		for (s = 0; s < SIDs.length; ++s) {
			if (typeof Sankey_Enrollment[SIDs[s]] === 'undefined') continue;
			//Get student s's class schedule for term t.
			var CurSched = Sankey_Enrollment[SIDs[s]][terms[t]];
			if (typeof CurSched === 'undefined') continue;
				
			for (c = 0; c < CurSched.length; ++c) {
				//These are used to correctly identify what students leave the major or transfer in

				if (typeof Sankey_ClassTotals[CurSched[c].catalog] == 'undefined')  {
					console.log("The course " + CurSched[c].catalog + " is referenced in the registration data but does not exist in the class data");
					continue;
				}
					
				//if (CurSched[c].grade == "NULL" && CurSched[c].grade) continue;

				Sankey_ClassTotals[CurSched[c].catalog][terms[t]].in.add(SIDs[s]);					//Class roster coming in
				Sankey_ClassTotals[CurSched[c].catalog][terms[t]].out.add(SIDs[s]);					//Class roster going out
				Sankey_ClassTotals[CurSched[c].catalog][terms[t]].dropouts.add(SIDs[s]);			//SIDs of students who did not take another class
				Sankey_ClassTotals[CurSched[c].catalog][terms[t]].grades[SIDs[s]] = CurSched[c].grade;//Save the grades for each student

				var Connections = new Set();
				if (prereq[Sankey_FilterBy+CurSched[c].catalog]) {
					prereq[Sankey_FilterBy+CurSched[c].catalog].forEach(function (d) { 
						Connections.add(d);
					});
				}
				var ConnectedOnce = false;
				var PastCheckAmount = Sankey_MaxPastChecking;

				++PastCheckAmount;

				//Get student s's class schedule for previous terms.
				for (pt = t-1; pt > -1 && ((!Sankey_OnlyPrereqLinks && !ConnectedOnce) || (Sankey_OnlyPrereqLinks && Connections.size != 0)); --pt) {
					var OldSched = Sankey_Enrollment[SIDs[s]][terms[pt]];
					if (typeof OldSched === 'undefined') continue;


					//Find out which prerequisite classes this student came from, and save them in an array
					for (i = 0; i < OldSched.length; ++i) {
						if (typeof Sankey_ClassTotals[OldSched[i].catalog] == 'undefined') continue;

						Sankey_ClassTotals[OldSched[i].catalog][terms[pt]].dropouts.delete(SIDs[s]);

						if (
							!Sankey_OnlyPrereqLinks 												//If we don't care about whether the preceding classes are prerequisites, add it as a 'camefrom'
							|| (Sankey_ShowRetakes && CurSched[c].catalog == OldSched[i].catalog) 	//If this is a retake, show it if deisred	
							|| (																	//otherwise
								prereq[Sankey_FilterBy+CurSched[c].catalog] 						//Check if the class has any prerequisites at all
								&& prereq[Sankey_FilterBy+CurSched[c].catalog].indexOf(Sankey_FilterBy+OldSched[i].catalog) != -1)	//if it does, is this preceding class one of them?
							) 
						{
							//Save this previous class as a 'camefrom' so we can draw a link from class 'i' in term 'pt' to the current class 'c' in term 't' for student 's'.
							Sankey_ClassTotals[CurSched[c].catalog][terms[t]].camefrom.push(
								{"sched":OldSched[i].catalog,
								"term":terms[pt], 
								"id":SIDs[s]
								}
							);
							ConnectedOnce = true;
							Connections.delete(Sankey_FilterBy+OldSched[i].catalog);
						}
					}
				}
			}
		}
	}


	//Reset the diagram to blank
	if (typeof Sankey_Cont != 'undefined')
		Sankey_Cont.remove();
			

	//Initialize the container
	Sankey_Cont = d3.select("body").append("div");
	var svg = Sankey_Cont.append("svg:svg");

	var Sankey_Width = (terms.length + .5) * Sankey_WPadding;
	var Sankey_Height = (classes.length + 3) * Sankey_HPadding;
	var ScrollbarWidth = getScrollBarWidth();

	if (Sankey_SidebarToggle) Sankey_Width += Sankey_WPadding*2;

	if (Sankey_Width < window.innerWidth) Sankey_Width = window.innerWidth;
	if (Sankey_Height < window.innerHeight) Sankey_Height = window.innerHeight;

	svg
		.attr("width", Sankey_Width)
		.attr("height", Sankey_Height)
	;

	
	var lines = svg.append("svg:g").attr("id","lines");				//A container for all the links
	var boxes = svg.append("svg:g").attr("id","boxes");				//A container for all the class boxes
	var LeftHeader = svg.append("svg:g").attr("id","LeftHeader");	//A container for all the left header labels and boxes
	var TopHeader = svg.append("svg:g").attr("id","TopHeader");		//A container for all the upper header labels and boxes
	var Background = svg.append("svg:g").attr("id","Background");	//A container for all the background boxes
	var Toolbar = svg.append("svg:g").attr("id","Toolbar");			//A container for all the toolbar elements
	var Sidebar = Sankey_Cont.append("div").attr("id","Sidebar");	//A container for all the sidebar elements
	var Bottombar = svg.append("svg:g").attr("id","Bottombar");		//A container for all the sidebar elements


	Toolbar.append("rect")													//Background coloring for the toolbar
		.attr("width", Sankey_Width)
		.attr("height",Sankey_HPadding)
		.attr("fill", "#AADDFF")
		.attr("fill-opacity", 1)
	;


	Bottombar.background = Bottombar.append("rect")							//Background coloring for the bottombar
		.attr("width", Sankey_Width)
		.attr("height",Sankey_HPadding)
		.attr("fill", "#AADDFF")
		.attr("fill-opacity", 1)
	;
	Bottombar.boxes = Bottombar.append('svg:g');


	var ToolbarButtons = Toolbar.append("svg:g").attr("id","ToolbarButtons");//A container for all the static button elements

	var SidebarActivateButton = Sidebar.append("div")
		.style("width",40)
		.style("height",40)
		.style("position","absolute")
		.style("text-align", "center")
		.style("border", "2px solid")
		.style("border-right", "none")
		.style("border-top-left-radius", "10px")
		.style("border-bottom-left-radius", "10px")
		.style("opacity", ".8")
		.attr("class", "ToolbarButton")
		.style("background-color", "#5CF")
		.on("click", function () {Sankey_SidebarToggle = !Sankey_SidebarToggle;GenerateSankey();})
	;
	SidebarActivateButton.append("text")
		.attr("class","Title")
		.style("font-size", "28pt")
		.text(function () {
			if (Sankey_SidebarToggle)
				return "X";
			return "+";
		}
	);


	if (Sankey_SidebarToggle) {
		
		Sidebar
			.style('display', 'block')
			.style("position","absolute")
			.attr("fill", "#AADDFF")
			.style("opacity", ".8")
			.style("background-color", "#AADDFF")
			.style("width", Sankey_WPadding*2)
			.style("height",window.innerHeight-(Sankey_HPadding*3) - ScrollbarWidth)

		;
		Sidebar.on("mouseover",function(d) {
				this.style["opacity"] = ".8";
			})
			.on("mouseout",function(d) {
				this.style["opacity"] = ".3";
			})
		;

		var SidebarTable = Sidebar.append("table").style("width","100%").style("background-color","#000000");

		var row1 = SidebarTable.append("tr").style("background-color","#EEF0FF");
		row1.append("td")
			.attr("class","Title")
			.attr("colspan",3)
			.text("Student Selector:")
		;

		var row2 = SidebarTable.append("tr").style("background-color","#FFFFFF");
		row2.append("td")
			.attr("class","Title")
			.style("width","50%")
			.text("Search: ")
		;
		var AddStudentInput = row2.append("td").append("input")
			.attr("id", "AddStudentInput")
			.attr("type","text")
			.attr("value", Sankey_LastSearchString)
			.style("width","100%")

		;
		document.getElementById('AddStudentInput').addEventListener('keyup', AddStudentSearch, false);

		row2.append("td")
			.append("input")
				.attr("type","button")
				.style("width","100%")
				.attr("class", "mybutton")
				.attr("value","Add")
				.on("click",TryAddStudent)
		;

		var row3 = SidebarTable.append("tr").style("background-color","#FFFFFF");
		var searchtable = row3.append("td").attr("colspan",3)
			.append("div")
				.style("height",150)
				.style("background-color", "#FEFEFE")
				.style("overflow-y","scroll")
				.style("padding","5px")
				.append('table').style("width","100%").style("background-color","#AAAAAA").attr("id", "searchtable");

		UpdateSearchTable();

		var row4 = SidebarTable.append("tr").style("background-color","#EEF0FF");
		row4.append("td")
			.attr("class","Title")
			.attr("colspan",3)
			.text("Selected Students: " + Sankey_SelectedStudents.size)
		;

		var row5 = SidebarTable.append("tr").style("background-color","#FFFFFF");
		var selection = row5.append("td")
			.attr("colspan",3)
			.append("div")
				.style("width",Sankey_WPadding*2 - ScrollbarWidth )
				.style("height",window.innerHeight-(Sankey_HPadding*8) - ScrollbarWidth)
				.style("background-color", "#FFFFFF")
				.style("overflow-y","scroll")
				.style("padding","5px")

		;

		selection.append("button")
			.attr("class", "mybutton")
			.style("float", "left")
			.style("margin-bottom", "5px")
			.text("Complement").on("click", function (d) {
				var tmp = new Set();
				Sankey_rawSIDs.forEach(function (d) {
					if (!Sankey_SelectedStudents.has(d)) Sankey_SelectedStudents.add(d);
					else Sankey_SelectedStudents.delete(d)
				});

				GenerateSankey();

				Tooltip();

			})
		;
		selection.append("button")
			.attr("class", "mybutton")
			.style("float", "right")
			.style("margin-bottom", "5px")
			.text("All Graduates")
			.on("click", function (d) {
				Sankey_rawGraduates.forEach(function(d) {
					Sankey_SelectedStudents.add(d);
				});
				GenerateSankey();

				Tooltip();
			})
		;

		selection.append("button")
			.attr("class", "mybutton")
			.style("float", "left")
			.style("margin-bottom", "5px")
			.text("Clear Selection").on("click", function (d) {
				Sankey_SelectedStudents.clear();
				GenerateSankey();

				Tooltip();

			})
		;
		selection.append("button")
			.attr("class", "mybutton")
			.style("float", "right")
			.style("margin-bottom", "5px")
			.text(function(){
				if (Sankey_FilterToggle) return "Don't filter";
				return "Filter by selection";
			})
			.on("click", function (d) {
				Sankey_FilterToggle = !Sankey_FilterToggle;
				GenerateSankey();
			})
		;


		var innertable = selection.append('table').style("width","100%").style("background-color","#AAAAAA");

		var TopRow = innertable.append("tr").style("background-color","#FFFFFF");
		TopRow.append('td').html("Student ID").attr("class","SubTitle").data(["id"]);
		TopRow.append('td').html("Name").attr("class","SubTitle").data(["name"]);

		var HoverColor = "#DDDDEE";
		var AltColor = "#F0F0FF";
		var SelectColor = "#99AAFF";

		var i = 0;
		Sankey_SelectedStudents.forEach(function(d) {
			var NewRow = innertable.append("tr");
			NewRow.data([{"d":d,"i":i}])
				.style("background-color",function(d) {

					if (d.i % 2) return AltColor;

					return "#FFFFFF";
				})
			;

			NewRow
				.on("click",function(d) {

					if (d.i % 2) this.style["background-color"] = AltColor;
					else this.style["background-color"] = "#FFFFFF";

					Sankey_SelectedStudents.delete(d.d);
					
					GenerateSankey();

					Tooltip();

				})
				.on("mouseover",function(d) {
					this.style["background-color"] = HoverColor;
				})
				.on("mouseout",function(d) {
					if (d.i % 2) this.style["background-color"] = AltColor;
					else this.style["background-color"] = "#FFFFFF";
				})
			;

			NewRow.append('td').html(d);

			NewRow.append('td').html(students[d].name);
			++i;
		});

	}

	var defs = svg.append("svg:defs").attr("id","definitions");		//A container for the gradient definitions
	var color = d3.scale.category20();

	var CullingSquare1 = svg.append("rect")							//A rect for removing the overlap between the top and left headers
		.attr("class", "Box")
		.attr("width", Sankey_WPadding/2)
		.attr("height", Sankey_HPadding)
		.attr("fill", "#AADDFF")
		.attr("fill-opacity", 1)
	;
	var CullingSquare2 = svg.append("rect")							//A rect for removing the overlap between the bottom and left headers
		.attr("class", "Box")
		.attr("width", Sankey_WPadding/2)
		.attr("height", Sankey_HPadding)
		.attr("fill", "#AADDFF")
		.attr("fill-opacity", 1)
			
	;


	//Define the gradients between classes. O(N^2) where N is the number of classes
	for (i = 0; i < classes.length; ++i) {
		for (j = 0; j < classes.length; ++j) {
			var gradient = defs.append("linearGradient")
					.attr("id", classes[i] + "_" + classes[j])
				;
				gradient.append("stop")
					.attr("offset", "0")
					.attr("stop-color", color(i))
				;
				gradient.append("stop")
					.attr("offset", "1")
					.attr("stop-color", color(j))
				;
		}
	}

	//create the leftmost vertical background grid
	LeftHeader.append("rect").attr("class", "Box")
		.attr("width", Sankey_WPadding)
		.attr("height", Sankey_Height)
		.attr("x", Sankey_WPadding * -.5).attr("y",Sankey_HPadding)
		.attr("fill", "#FFBBAA")
		.attr("fill-opacity", 1)
	;

	//Create the topmost horizontal background grid
	TopHeader.append("rect").attr("class", "Box")
		.attr("width", Sankey_Width)
		.attr("height", Sankey_HPadding)
		.attr("x", Sankey_WPadding/2).attr("y", Sankey_HPadding)
		.attr("fill", "#AADDFF")
		.attr("fill-opacity", 1)
	;
			
	//Create the term label headers and inner vertical background grid
	for (i = 0; i < terms.length; ++i) {
		TopHeader.append("svg:text")
			.attr("stroke", "#000000")
			.attr("class", "Title")
			.attr("transform", "translate(" + ((i+1) * Sankey_WPadding) + "," + 3*Sankey_HPadding/2 + ")")
			.text(
				function() {
					if (terms[i][3] == '3')
						return "Spring " + terms[i].substr(1,2);
					else return "Fall " + terms[i].substr(1,2);
				}
			).attr("text-anchor", "middle")
		;
		if (i % 2) {
			Background.append("rect").attr("class", "Box")
				.attr("width", Sankey_WPadding)
				.attr("height", Sankey_HPadding * (classes.length+1))
				.attr("x", Sankey_WPadding * (i+.5)).attr("y", 0)
				.attr("fill", function () {return color(3);})
				.attr("fill-opacity", Sankey_GridOpacity)
			;
		}
	}

	//create the class label headers and inner horizontal background grid
	for (j = 0; j < classes.length; ++j) {
		LeftHeader.append("svg:text")
			.attr("stroke", "#000000")
			.attr("class", "Title")
			.attr("transform", "translate(" + Sankey_WPadding/8 + "," + ((j+2.5) * Sankey_HPadding) + ")")
			.text(classes[j]).attr("dominant-baseline", "central")
		;

		Background.append("rect").attr("class", "Box")
			.attr("width", Sankey_WPadding * (terms.length+.5))
			.attr("height", Sankey_HPadding)
			.attr("x", 0).attr("y", Sankey_HPadding * (j+2))
			.attr("fill", function () {return color(j%2);})
			.attr("fill-opacity", Sankey_GridOpacity)
		;
	}

	//Add some buttons to the toolbar

	//Prerequisite classes
	ToolbarButtons.append("rect").attr("width",190).attr("height",20).attr("class", "ToolbarButton")
		.attr("fill", "#FBA")
		.on("click", function () {Sankey_OnlyPrereqClasses = !Sankey_OnlyPrereqClasses;GenerateSankey();})
	;
	ToolbarButtons.append("svg:text").attr("class","ButtonText").text(function () {
		if (Sankey_OnlyPrereqClasses)
			return "Show Non-prerequisite Classes";
		return "Hide Non-prerequisite Classes";
	});


	//Serial vs prereq links
	ToolbarButtons.append("rect").attr("width",190).attr("height",20).attr("class", "ToolbarButton")
		.attr("fill", "#FBA")
		.on("click", function () {Sankey_OnlyPrereqLinks = !Sankey_OnlyPrereqLinks;GenerateSankey();})
	;
	ToolbarButtons.append("svg:text").attr("class","ButtonText").text(function () {
		if (Sankey_OnlyPrereqLinks)
			return "Serial Links";
		return "Prerequisite Links";
	});

	if (Sankey_OnlyPrereqLinks) {
		ToolbarButtons.append("rect").attr("width",190).attr("height",20).attr("class", "ToolbarButton")
			.attr("fill", "#5CF")
			.on("click", function () {Sankey_ShowRetakes = !Sankey_ShowRetakes;GenerateSankey();})
		;
		ToolbarButtons.append("svg:text").attr("class","ButtonText").text(function () {
			if (Sankey_ShowRetakes)
				return "Hide Retakes";
			return "Show Retakes";
		});
	}

	//Make the above headers and toolbars follow the users scrolling
	window.onscroll = function () {
		var ScrollbarWidth = getScrollBarWidth();
		LeftHeader.selectAll("text").each(function (d,j) {
			this.setAttribute("transform", "translate(" + ((Sankey_WPadding/8) + window.scrollX) + "," + ((j+2.5) * Sankey_HPadding) + ")");
		});
		LeftHeader.selectAll("rect").each(function () {
			this.setAttribute("x", (Sankey_WPadding * -.5) + window.scrollX);
		});
		TopHeader.selectAll("text").each(function (d,i) {
			this.setAttribute("transform", "translate(" + ((i+1) * Sankey_WPadding) + "," + (3*Sankey_HPadding/2 + window.scrollY) + ")");
		});
		TopHeader.selectAll("rect").each(function () {
			this.setAttribute("y", Sankey_HPadding+window.scrollY);
		});

		Toolbar.selectAll("rect").each(function (d,i) {
			this.setAttribute("x", window.scrollX);
			this.setAttribute("y", window.scrollY);
		});
		Bottombar.background.attr("y", window.scrollY+window.innerHeight-Sankey_HPadding - ScrollbarWidth);
		Bottombar.boxes.selectAll("rect").each(function (d,i) {
			this.setAttribute("y", window.scrollY+window.innerHeight - Sankey_HPadding*(1-Sankey_NodeHeight)/2 + (Sankey_NodeHeight - d.value) - ScrollbarWidth - d.offset);
		});

		var tmp = (window.scrollX+window.innerWidth) - ScrollbarWidth;
		if (Sankey_SidebarToggle) {
			tmp -= Sankey_WPadding*2;
			if (tmp > Sankey_Width - (Sankey_WPadding*2)) tmp = Sankey_Width - (Sankey_WPadding*2);
			SidebarActivateButton.style("left",-40);
		}
		else {
			SidebarActivateButton.style("left",tmp-42);
		}
		Sidebar.style("left", tmp);


		var tmp = (window.scrollY+(Sankey_HPadding*2));
		if (tmp > Sankey_Height - window.innerHeight + (Sankey_HPadding*2) + ScrollbarWidth) tmp = Sankey_Height - window.innerHeight + (Sankey_HPadding*2) + ScrollbarWidth;

		Sidebar.style("top", tmp);
		if (!Sankey_SidebarToggle) {
			SidebarActivateButton.style("top",tmp);
		}
		else SidebarActivateButton.style("top",0);



		ToolbarButtons.selectAll("rect").each(function (d,i) {
			this.setAttribute("x", Sankey_WPadding/4 + (i * 200) + window.scrollX);
			this.setAttribute("y", Sankey_HPadding/2 + window.scrollY);
		});
		ToolbarButtons.selectAll("text").each(function (d,i) {
			this.setAttribute("x", Sankey_WPadding/4 + ((i+.5) * 200) + window.scrollX);
			this.setAttribute("y", 3*Sankey_HPadding/4 + window.scrollY);
		});

		CullingSquare1.attr("x", (window.scrollX));
		CullingSquare1.attr("y", Sankey_HPadding + (window.scrollY));

		CullingSquare2.attr("x", (window.scrollX));
		CullingSquare2.attr("y", (window.scrollY) + innerHeight - (Sankey_HPadding) - ScrollbarWidth);

	};




	var LargestClass = 0;
	var LargestTerm = 0;
	var offset = Sankey_WPadding/2 * Sankey_NodeWidth;
	var ConnectorData = [];		//Holds the data for how the links should be drawn
	var SourceData = [];		//Holds data concerning the placement of the left ends of the links
	var nodes = [];				//Holds the box-shaped nodes so that they can be resized later
	var TermTotals = [];

	//Generate the diagram's inner data
	for (i = 0; i < terms.length; ++i) {

		var graduates,nongraduates;
		graduates = new Set();
		nongraduates = new Set();

		for (j = 0; j < classes.length; ++j) {
			var data = Sankey_ClassTotals[classes[j]][terms[i]];
			data.sum = data.in.size;

			if (data.sum == 0) continue;	//If there is noone enrolled, skip this class

			if (data.sum > LargestClass) LargestClass = data.sum;

			ConnectorData[(i*classes.length) + j] = [];

			//Create a container for the data
			var node = boxes.append("svg:g")
				.attr("transform", "translate(" + ((i+1) * Sankey_WPadding) + "," + ((j+1.5) * Sankey_HPadding) + ")");

			//Create a rectangle
			nodes[i*classes.length + j] = node.append("rect")
				.attr("width", Sankey_WPadding * Sankey_NodeWidth)
				.attr("x", -Sankey_WPadding/2 * Sankey_NodeWidth)
				.attr("stroke", function () {return color(j);})
				.attr("fill", function () {return color(j);})
				.attr("fill-opacity", .6)
				.attr("stroke-opacity", 0)
			;

			//Show all students when clicked
			var tmp = [];
			data.in.forEach(function (d) {
				var S = students[d];
				tmp.push({"id":d,"name":S.name,"grade":data.grades[d], "gender":S.gender,"ethnicity":S.ethnicity});

				if (S.finalterm == terms[i]) {
					if (S.graduate) graduates.add(d);
					else nongraduates.add(d);
				}
			});


			if (terms[i][3] == '3')
				tmp.term = "Spring " + terms[i].substr(1,2);
			else tmp.term = "Fall " + terms[i].substr(1,2);

			tmp.name = Sankey_FilterBy + classes[j] + ", " + tmp.term;

			nodes[i*classes.length + j].data([tmp])
				.on("click", function (d) {
					Tooltip(d);
				})
			;


			//Create a text label for the class
			if (Sankey_ShowNumbers)
				node.append("svg:text").text(data.sum).attr("text-anchor", "middle").attr("dominant-baseline", "central");

			//Sort the 'camefrom's to count duplicates properly
			data.camefrom.sort(
				function (a,b) {
					if (a.sched < b.sched) return -1;
					else if (a.sched == b.sched) {
						if (a.term < b.term) return -1;
						else if (a.term < b.term) return 0;
						return 1;
					}
					else return 1;
				}
			);	
			 
			var summed = [];		//Holds a summed version of 'camefrom' without duplicates
			var count = -1;
			var last = "";
			var uniqueids = [];
			var total = 0;			//Represents the enrollment total with overcounting

			//Sum the 'camefrom' data without overcounting
			for (k = 0; k < data.camefrom.length; ++k) {

				Sankey_ClassTotals[data.camefrom[k].sched][data.camefrom[k].term].out.delete(data.camefrom[k].id);
				data.in.delete(data.camefrom[k].id);

				if (data.camefrom[k].sched + data.camefrom[k].term != last) {
					last = data.camefrom[k].sched + data.camefrom[k].term;

					if (count != -1) total += summed[count].sum;

					summed.push({"name": data.camefrom[k].sched, "term":data.camefrom[k].term,"sum" : 1, "ids": [data.camefrom[k].id]});
					uniqueids = [];
					uniqueids.push(data.camefrom[k].id);

					++count;
				}

				else if (uniqueids.indexOf(data.camefrom[k].id) == -1) {
					uniqueids.push(data.camefrom[k].id);
					summed[count].ids.push(data.camefrom[k].id);
					++summed[count].sum;
				}
			}

			if (count != -1) total += summed[count].sum;

			var totalWidth = 0;
			//Use the summed camefrom data to create the links to previous classes
			for (k = 0; k < summed.length; ++k) {

				var x = terms.indexOf(summed[k].term);	//x-index of camefrom class
				var y = classes.indexOf(summed[k].name);//y-index of camefrom class

				var width = (summed[k].sum / total);	//Find the size of this link relative to it's siblings


				if (typeof SourceData[(x*classes.length) + y] === 'undefined') 
					SourceData[(x*classes.length) + y] = {"Total" : summed[k].sum,"Used":0};
				
				else SourceData[(x*classes.length) + y].Total += summed[k].sum;

				ConnectorData[(i*classes.length) + j].push({
					"x1":(((x+1) * Sankey_WPadding) + offset),	"y1":((y+2.5) * Sankey_HPadding),
					"x2":(((i+1) * Sankey_WPadding) - offset),	"y2":((j+2.5) * Sankey_HPadding),
					"Rwidth":width/2, "sum":summed[k].sum, "x": x, "y":y , "TotalUsed":totalWidth/ total, "ids":summed[k].ids
				});
				totalWidth+= summed[k].sum;
			}
		}
		TermTotals[terms[i]] = {"graduates":graduates,"nongraduates":nongraduates};
		if (LargestTerm < graduates.size + nongraduates.size) LargestTerm = graduates.size + nongraduates.size;
		
	}

	//Use the gathered data to form and adjust the links properly.
	for (y = 0; y < classes.length; ++y) {
		for (x = 0; x < terms.length; ++x) {

			var data = ConnectorData[x*classes.length + y];

			if (typeof data === 'undefined') continue;

			//Find the height of this node (the rightmost one in terms of connections)
			var RTotals = Sankey_ClassTotals[classes[y]][terms[x]];

			var Rused = ((RTotals.sum - RTotals.in.size) / RTotals.sum);	//The percentange of students in this cource who came from a valid link
			var RsquareWidth = (RTotals.sum / LargestClass) * Sankey_HPadding * Sankey_NodeHeight;

			//Set the final height of this node
			nodes[x*classes.length + y].attr("height", RsquareWidth)
				.attr("y", Sankey_HPadding + (Sankey_NodeHeight - RsquareWidth)/2)
			;

			//Get a user-friendly format for the term
			var rate = 1-((RTotals.dropouts.size*1.0) / (RTotals.sum*1.0));
			var termname = "fall";
			if (terms[x][3] == '7')
				termname = "spring";
			termname += " " + terms[x].substr(1,2);


			//Add a tooltip
			var tooltipTxt = Sankey_FilterBy + " " + classes[y] + " has " + RTotals.sum + " students during the " + termname + " term";
			if (Sankey_OnlyPrereqClasses)
				tooltipTxt += "\n" + Math.floor(rate*100) + "% take another 'prerequisite' class";

			else tooltipTxt += "\n" + Math.floor(rate*100) + "% take another " + Sankey_FilterBy + " class";

			nodes[x*classes.length + y].append("svg:title")
				.text(tooltipTxt)
			;

			for (k = 0; k < data.length; ++k) {
				var d = data[k];
				var source = SourceData[(d.x * classes.length) + d.y];

				var LTotals = Sankey_ClassTotals[classes[d.y]][terms[d.x]];
				var Lused = ((LTotals.sum - LTotals.out.size) / LTotals.sum);	//The percentage of students in the source course who continued on to a valid course

				//Find the height of the connected node (the leftmost one)
				var LsquareWidth = (LTotals.sum / LargestClass) * Sankey_HPadding * Sankey_NodeHeight;

				var Lwidth = ((d.sum / source.Total) * LsquareWidth/2) * Lused;	//The width of the left connection
				d.Rwidth *= RsquareWidth * Rused;								//The width of the right connection

				//Offsets for knowing exactly where to connect the links to the rectangle so they don't overlap
				var LLinkOffset = ((Sankey_NodeHeight - LsquareWidth)/2 + (Lwidth) + ((source.Used/source.Total) * LsquareWidth)*Lused);
				var RLinkOffset = ((Sankey_NodeHeight - RsquareWidth)/2 + (d.Rwidth) + (d.TotalUsed * RsquareWidth)*Rused);



				d["y1"] += LLinkOffset;
				d["y2"] += RLinkOffset;


	          	var mid = d["x1"] + (d["x2"] - d["x1"])/2;			//Get the midpoint of the start and end


	          	//Generate the final path data
				var path = 
					"M"  											//Start
					+ " " + d["x1"] + " " + (d["y1"]-Lwidth)		//Left top coord

					+ " C"											//Draw a curve
					+ " " + mid 	+ " " + (d["y1"]-Lwidth)		//Left top control point
					+ " " + mid 	+ " " + (d["y2"]-d.Rwidth)		//Right top control point
					+ " " + d["x2"] + " " + (d["y2"]-d.Rwidth)		//Right top coord

					+ " L"											//Draw a line
					+ " " + d["x2"] + " " + (d["y2"]+d.Rwidth)		//Right bottom coord

					+ " C"											//Draw a curve
					+ " " + mid 	+ " " + (d["y2"]+d.Rwidth)		//Right bottom control point
					+ " " + mid 	+ " " + (d["y1"]+Lwidth)		//Left bottom control point
					+ " " + d["x1"] + " " + (d["y1"]+Lwidth)		//Left bottom coord

					+ " Z"											//Stop

				source.Used += d.sum;

				//Create the d3 path
				var link = lines.append("svg:path")
					.attr("class", "link")
					.attr("d", path)
					.attr("stroke-width", 0)
					.style("fill", "url(#" + (classes[d.y]) + "_" + (classes[y]) + ")");
				;
				//Add a tooltip
				link.append("svg:title")
					.text(d.sum + " from " + Sankey_FilterBy + " "  + classes[d.y] + " to " + Sankey_FilterBy + " " + classes[y])
				;

				//Show extra data when clicked
				var tmp = [];
				d.ids.forEach(function (d) {
					var S = students[d];
					tmp.push({"id":d,"name":S.name,"grade":LTotals.grades[d], "gender":S.gender,"ethnicity":S.ethnicity});
				});


				if (terms[d.x][3] == '3')
					tmp.term = "Spring " + terms[d.x].substr(1,2);
				else tmp.term = "Fall " + terms[d.x].substr(1,2);

				tmp.term += " to ";

				if (terms[x][3] == '3')
					tmp.term += "Spring " + terms[x].substr(1,2);
				else tmp.term += "Fall " + terms[x].substr(1,2);

				tmp.name = Sankey_FilterBy + " " + classes[d.y] + " to " + Sankey_FilterBy + " " + classes[y] + ", " + tmp.term;

				link.data([tmp])
					.on("click", function (d) {
						Tooltip(d);
					})
				;
			}
		}
	}

	//Create the bottom bar to show term totals
	for (x = 0; x < terms.length; ++x) {
		//Show extra data when clicked
		var tmp = [];

		//Get the combination of all students finishing in term x
		TermTotals[terms[x]].graduates.forEach(function (d) {
			var S = students[d];
			tmp.push({"id":d,"name":S.name,"grade":null, "gender":S.gender,"ethnicity":S.ethnicity});
		});
		TermTotals[terms[x]].nongraduates.forEach(function (d) {
			var S = students[d];
			tmp.push({"id":d,"name":S.name,"grade":null, "gender":S.gender,"ethnicity":S.ethnicity});
		});

		if (terms[x][3] == '3')
			tmp.term = "Spring " + terms[x].substr(1,2);
		else tmp.term = "Fall " + terms[x].substr(1,2);

		tmp.name = "Students who discontinue in the " + tmp.term + " term";

		var totalstudents = TermTotals[terms[x]].nongraduates.size + TermTotals[terms[x]].graduates.size;
		var tooltiptxt = totalstudents + " students finish in " + tmp.term + "\r\n" 
			+ TermTotals[terms[x]].graduates.size + " graduate, " + TermTotals[terms[x]].nongraduates.size + " don't";

		nongrads = ((TermTotals[terms[x]].nongraduates.size) / LargestTerm) * Sankey_NodeHeight * Sankey_HPadding;
		grads = ((TermTotals[terms[x]].graduates.size) / LargestTerm) * Sankey_NodeHeight * Sankey_HPadding;



		Bottombar.boxes
				.append("rect")
				.data([{"value":Sankey_NodeHeight * Sankey_HPadding,"offset":0,"d":tmp}])
				.attr("width", Sankey_WPadding * Sankey_NodeWidth)
				.attr("height", function(d){return d.value;})
				.attr("x", ((x+1) * Sankey_WPadding)-Sankey_WPadding/2 * Sankey_NodeWidth)
				.attr("stroke", "#FFFFFF")
				.attr("fill", "#FFFFFF")

				.on("click", function (d) {
					Tooltip(d.d);

					BumpUpTooltip();
				})
				.append("svg:title").text(tooltiptxt)
		;

		Bottombar.boxes
				.append("rect")
				.data([{"value":nongrads,"offset":0,"d":tmp}])
				.attr("width", Sankey_WPadding * Sankey_NodeWidth)
				.attr("height", function(d){return d.value;})
				.attr("x", ((x+1) * Sankey_WPadding)-Sankey_WPadding/2 * Sankey_NodeWidth)
				.attr("stroke", "#AAAAAA")
				.attr("fill", "#AAAAAA")

				.on("click", function (d) {
					Tooltip(d.d);
					BumpUpTooltip();
				})
				.append("svg:title").text(tooltiptxt)
		;

		Bottombar.boxes
				.append("rect")
				.data([{"value":grads,"offset":nongrads,"d":tmp}])
				.attr("width", Sankey_WPadding * Sankey_NodeWidth)
				.attr("height", function(d){return d.value;})
				.attr("x", ((x+1) * Sankey_WPadding)-Sankey_WPadding/2 * Sankey_NodeWidth)
				.attr("stroke", "#FF8800")
				.attr("fill", "#FF8800")

				.on("click", function (d) {
					Tooltip(d.d);
					BumpUpTooltip();
				})
				.append("svg:title").text(tooltiptxt)
		;


	}

	window.onscroll();			//Call the scroll function we just made to position the elements the lazy way
}
function BumpUpTooltip() {
	var div = document.getElementById('tooltip');
	if (!div) return;
	var Pos = {"x":0,"y":0};
	Pos.x = Number(div.style.left.substr(0,div.style.left.length-2));
	Pos.y = Number(div.style.top.substr(0,div.style.top.length-2));

	div.style.position = 'absolute';
	div.style.top = (Pos.y - 500)  + 'px';

	Sankey_TooltipInitialPos.y -= 500;
}
function FilterPrereqs(Data) {
	//Turn the prerequisite data into a format that is more helpful for establishing links
	//between classes.
	 
	var prereqs = [];
	var nodes = new Set();
	Data.forEach(
		function (d) {
			if (d["type"] == "P" && d["subject"] == Sankey_FilterBy && d["req_subject"] == Sankey_FilterBy) {
				if (typeof prereqs[d["subject"]+d["catalog"]] === 'undefined')
					prereqs[d["subject"]+d["catalog"]] = [];
				prereqs[d["subject"]+d["catalog"]].push(d["req_subject"]+d["req_catalog"]);

				nodes.add(d["subject"]+d["catalog"]);
				nodes.add(d["req_subject"]+d["req_catalog"]);

			}
		}
	);
	prereqs["nodes"] = Array.from(nodes);
	return prereqs;
}


function Tooltip(SD,SortBy) {
	//This function creates an info window that allows the user to select students to filter by
	//and shows relevant info in pie charts

	//Save old tooltip data in case we need it
	var div = document.getElementById('tooltip');
	var x,y;
	if (typeof div != 'undefined' && div != null) {
		x = div.style.left;
		y = div.style.top;
	}

	//Clear the old tooltip to make foom for this new one
	d3.select('.ourTooltip').remove();

	if (typeof SD === 'undefined' || SD == null) {
		SD = Sankey_CurrentTooltipData;
		SortBy = "name";
	}
	else {
		Sankey_CurrentTooltipData = SD;
	}

	if (typeof SD === 'undefined' || SD == null) return;

	var base = d3.select('body').append('div').attr("class", "ourTooltip").attr("id","tooltip");

	base
		.style('display', 'block')
		.style("position","absolute")
		.data([SD])
	;

    window.addEventListener('mouseup', mouseUp, false);	//Add the stop-drag handler

    //If this is the first time generating this window, position it anew
    if (typeof SortBy === 'undefined') {
    	base
    		.style('top', (d3.event.pageY) + 'px')
			.style('left', (d3.event.pageX - 350) + 'px')
			.style("opacity", 0)
			.transition().delay(250).duration(250).style('opacity', .8)
		;
    	mouseUp();			//The lazy way to position the window		
    	SortBy = "name";
	}		
	//We don't need to run the fade-in if this is a sorting event
	else {
		base
			.style('top', y)
			.style('left', x)
			.style("opacity", .8)
		;
	}	

	SD.sort(function (a,b) {
		if (a[SortBy] < b[SortBy]) return -1;
		else if (a[SortBy] == b[SortBy]) 0;
		return 1;
	});

	base.on("mouseover",function() {
			this.style["opacity"] = ".8";
		})
		.on("mouseout",function() {
			this.style["opacity"] = ".3";
		})
	;


	var table = base.append('table').style("background-color","#000000");
	var row1 = table.append("tr").attr("id","tooltip-toolbar").attr("height",30).style("background-color","#EEF0FF");	//The title bar for the tooltip window

	//Add a listener to allow the toolbar to be dragged
	document.getElementById('tooltip-toolbar').addEventListener('mousedown', mouseDown, false);	

	//Window title
	var titlebar = row1.append("td").append("div");
	titlebar.append("text")
		.attr("class", "Title")
		.style("padding-left","10px")
		.text(SD.name)
	;
	titlebar.append("text")
		.attr("class", "Title")
		.style("padding-right","10px")
		.style("float","right")
		.text(SD.length + " students")
	;

	//Close window button
	row1.append("td").append("button")
		.style("float", "right")
		.style("margin-right", "10px")
		.text("Close").on("click", function () {d3.select('.ourTooltip').remove();Sankey_CurrentTooltipData = null;})
	;


	var row1,row2,row3,selection;

	row1 = table.append("tr").style("background-color","#FFFFFF");
	row2 = table.append("tr").style("background-color","#FFFFFF");
	row3 = table.append("tr").style("background-color","#FFFFFF");

	selection = row1.append("td").append("div")
		.style("width",500)
		.style("padding","10px")
	;

	selection.append("button")
		.attr("class", "mybutton")
		.style("float", "left")
		.style("margin-bottom", "5px")
		.text("Complement").on("click", function (d) {
			var tmp = new Set();

			d.forEach(function(d) {
				if (Sankey_SelectedStudents.has(d.id)) Sankey_SelectedStudents.delete(d.id);
				else Sankey_SelectedStudents.add(d.id);
			});

			GenerateSankey();

			Tooltip();

		})
	;
	selection.append("button")
		.attr("class", "mybutton")
		.style("float", "right")
		.style("margin-bottom", "5px")
		.text("All Graduates")
		.on("click", function (d) {
			d.forEach(function(d) {
				if (Sankey_rawGraduates.indexOf(d.id) != -1)
					Sankey_SelectedStudents.add(d.id);
			});
			GenerateSankey();

			Tooltip();
		})
	;

	selection = row2.append("td").append("div")
		.style("width",500)
		.style("padding","10px")
	;

	selection.append("button")
		.attr("class", "mybutton")
		.style("float", "left")
		.style("margin-bottom", "5px")
		.text("Clear Selection").on("click", function (d) {
			d.forEach(function(id) {
				Sankey_SelectedStudents.delete(id.id);
			});
			GenerateSankey();

			Tooltip(null,SortBy);

		})
	;
	selection.append("button")
		.attr("class", "mybutton")
		.style("float", "right")
		.style("margin-bottom", "5px")
		.text("Select All").on("click", function (d) {
			d.forEach(function(id) {
				Sankey_SelectedStudents.add(id.id);
			});
			GenerateSankey();

			Tooltip(null,SortBy);

		})
	;

	selection = row3.append("td").append("div")
		.style("width",500)
		.style("height",700)
		.style("overflow-y","scroll")
		.style("padding","10px")
	;
	var innertable = selection.append('table').style("background-color","#AAAAAA").style("width","100%");


	var TopRow = innertable.append("tr").style("background-color","#FFFFFF");;
	TopRow.append('td').html("Student ID").attr("class","SubTitle").data(["id"]);
	TopRow.append('td').html("Name").attr("class","SubTitle").data(["name"]);
	//TopRow.append('td').html("Gender").attr("class","SubTitle").data(["gender"]);
	//TopRow.append('td').html("Ethnicity").attr("class","SubTitle").data(["ethnicity"]);
	TopRow.append('td').html("Grade").attr("class","SubTitle").data(["grade"]);
	var HoverColor = "#DDDDEE";
	TopRow.selectAll('td')
		.each(function (d) {
			var sortby = d;
			this.onclick = function () {
				
				var div = document.getElementById('tooltip');
				var x = div.style.left;
				var y = div.style.top;

				Tooltip(SD,sortby);

				div = document.getElementById('tooltip');

				div.style.left = x;
				div.style.top = y;

			};
			this.onmouseover = function () {
				this.style["background-color"] = HoverColor;
			}
			this.onmouseout = function () {
				this.style["background-color"] = "#FFFFFF";
			}
		})
	;

	var AltColor = "#F0F0FF";
	var SelectColor = "#99AAFF";
	
	//Create a list of students in this class/link and make them selectable
	SD.forEach(function(d,i) {
		var NewRow = innertable.append("tr");

		NewRow.data([d.id])
			.style("background-color",function(d) {
				if (Sankey_SelectedStudents.has(d))
					return SelectColor;

				if (i % 2) return AltColor;

				return "#FFFFFF";
			})
		;

		NewRow
			.on("click",function(d) {
				if (Sankey_SelectedStudents.has(d)) {
					if (i % 2) this.style["background-color"] = AltColor;
					else this.style["background-color"] = "#FFFFFF";

					Sankey_SelectedStudents.delete(d);
				}
				else {
					this.style["background-color"] = SelectColor;
					Sankey_SelectedStudents.add(d);
				}

				GenerateSankey();
			})
			.on("mouseover",function(d) {
				this.style["background-color"] = HoverColor;
			})
			.on("mouseout",function(d) {
				if (Sankey_SelectedStudents.has(d)) {
					this.style["background-color"] = SelectColor;

				}
				else {
					
					if (i % 2) this.style["background-color"] = AltColor;
					else this.style["background-color"] = "#FFFFFF";
				}
			})
		;

		NewRow.append('td').html(d.id);

		NewRow.append('td').html(d.name);
		/*
		if (d.gender == 'M')
			NewRow.append('td').html("Male");
		else NewRow.append('td').html("Female");
		*/
		//NewRow.append('td').html(d.ethnicity.toLowerCase());

		NewRow.append('td').html(d.grade);
	});


	selection = row1.append("td").attr("height","100%").style("vertical-align", "top").attr("rowspan",3);

	var Charts = selection.append("svg:svg").attr("width","200px").attr("height","100%");

	var radius = 50;
	var padding = 50;

	//Titles for the charts
	Charts.append("svg:text")
		.attr("transform", "translate(" + (radius + padding) + "," + padding/2 + ")")
		.attr("text-anchor","middle")
		.text("Grades")
	;
	Charts.append("svg:text")
		.attr("transform", "translate(" + (radius + padding) + "," + ((radius + padding)*2 + padding) + ")")
		.attr("text-anchor","middle")
		.text("Gender")
	;
	Charts.append("svg:text")
		.attr("transform", "translate(" + (radius + padding) + "," + ((radius + padding)*4 + 3*padding/2) + ")")
		.attr("text-anchor","middle")
		.text("Ethnicity")
	;

	//Format the data for charting and create their respective containers
	var GradeCont = Charts.append("svg:g").attr("transform", "translate(" + (radius + padding) + "," + (radius + padding + padding/2) + ")");
	var Grades = [
		"NULL",
		"*B-",
		"*C+",
		"A",
		"A-",
		"B+",
		"B",
		"B-",
		"C+",
		"C",
		"C-",
		"CR",
		"D+",
		"D",
		"D-",
		"F",
		"I",
		"IC",
		"NC",
		"RP",
		"RPT",
		"W",
		"WU"
	];
	var GradeData = [
		{"id":"NULL","value":0},
		{"id":"*B-","value":0},
		{"id":"*C+","value":0},
		{"id":"A","value":0},
		{"id":"A-","value":0},
		{"id":"B+","value":0},
		{"id":"B","value":0},
		{"id":"B-","value":0},
		{"id":"C+","value":0},
		{"id":"C","value":0},
		{"id":"C-","value":0},
		{"id":"CR","value":0},
		{"id":"D+","value":0},
		{"id":"D","value":0},
		{"id":"D-","value":0},
		{"id":"F","value":0},
		{"id":"I","value":0},
		{"id":"IC","value":0},
		{"id":"NC","value":0},
		{"id":"RP","value":0},
		{"id":"RPT","value":0},
		{"id":"W","value":0},
		{"id":"WU","value":0}
	];
	var GenderCont = Charts.append("svg:g").attr("transform", "translate(" + (radius + padding) + "," + ((radius + padding)*3 + padding) + ")");
	var GenderData = [{"id":"Male","value":0},{"id":"Female","value":0}];
	var EthnicityCont = Charts.append("svg:g").attr("transform", "translate(" + (radius + padding) + "," + ((radius + padding)*5 + 3*padding/2) + ")");
	var Ethnicities = [
		"WHITE",
		"NULL",
		"DECLINE",
		"MEXAMER",
		"ASIAN",
		"EUROPEAN",
		"HISPA",
		"TWOMORE",
		"CENTAMER",
		"NORESPON",
		"FILIPINO",
		"BLACK",
		"CHINESE",
		"MEXICAN",
		"AMIND",
		"OTHRNWT",
		"UNKNOWN",
		"HISPANIC",
		"BLACKPRF",
		"OTHRLTHS",
		"SOAMER",
		"OTHRWHT",
		"CUBAN",
		"LAOTIAN",
		"OTHRSEAS",
		"PUERTOR",
		"OTHRPACI",
		"JAPANESE",
		"SAMOAN",
		"PACIF",
		"KOREAN",
		"CAMBOD",
		"ASINDIAN",
		"OTHRASN",
		"HAWAIIAN",
		"VIETNAME",
		"AFRAM",
		"GUAMANIA",
		"THAI",
		"DECHISPA",
		"OTHER",
		"OTHRAMIN",
		"HMONG"
	];
	var EthnicityData = [
		{"id":"WHITE","value":0},
		{"id":"NULL","value":0},
		{"id":"DECLINE","value":0},
		{"id":"MEXAMER","value":0},
		{"id":"ASIAN","value":0},
		{"id":"EUROPEAN","value":0},
		{"id":"HISPA","value":0},
		{"id":"TWOMORE","value":0},
		{"id":"CENTAMER","value":0},
		{"id":"NORESPON","value":0},
		{"id":"FILIPINO","value":0},
		{"id":"BLACK","value":0},
		{"id":"CHINESE","value":0},
		{"id":"MEXICAN","value":0},
		{"id":"AMIND","value":0},
		{"id":"OTHRNWT","value":0},
		{"id":"UNKNOWN","value":0},
		{"id":"HISPANIC","value":0},
		{"id":"BLACKPRF","value":0},
		{"id":"OTHRLTHS","value":0},
		{"id":"SOAMER","value":0},
		{"id":"OTHRWHT","value":0},
		{"id":"CUBAN","value":0},
		{"id":"LAOTIAN","value":0},
		{"id":"OTHRSEAS","value":0},
		{"id":"PUERTOR","value":0},
		{"id":"OTHRPACI","value":0},
		{"id":"JAPANESE","value":0},
		{"id":"SAMOAN","value":0},
		{"id":"PACIF","value":0},
		{"id":"KOREAN","value":0},
		{"id":"CAMBOD","value":0},
		{"id":"ASINDIAN","value":0},
		{"id":"OTHRASN","value":0},
		{"id":"HAWAIIAN","value":0},
		{"id":"VIETNAME","value":0},
		{"id":"AFRAM","value":0},
		{"id":"GUAMANIA","value":0},
		{"id":"THAI","value":0},
		{"id":"DECHISPA","value":0},
		{"id":"OTHER","value":0},
		{"id":"OTHRAMIN","value":0},
		{"id":"HMONG","value":0}
	];

	//Processing data into a chartable form
	SD.forEach(function(d) {
		if (d.gender == 'M')
			++GenderData[0].value;
		else ++GenderData[1].value;

		var index = Ethnicities.indexOf(d.ethnicity);
		if (index != -1) 
			++EthnicityData[index].value;

		index = Grades.indexOf(d.grade);
		if (index != -1) 
			++GradeData[index].value;
		
	});

	//Rid ourselves of empty data points
	GradeData = GradeData.filter(function (d) {
		if (d.value != 0) return d;
		return;
	});
	GenderData = GenderData.filter(function (d) {
		if (d.value != 0) return d;
		return;
	});
	EthnicityData = EthnicityData.filter(function (d) {
		if (d.value != 0) return d;
		return;
	});

	//Use the generated data to make the pies
	MakePie(GradeCont,GradeData,radius,padding);
	MakePie(GenderCont,GenderData,radius,padding);
	MakePie(EthnicityCont,EthnicityData,radius,padding);


	return base;
}
function MakePie(Container,data,radius,padding) {
	//Creates a pie chart in 'container'

	Container.data([data]);
	var color = d3.scale.category20();
	var pie = d3.layout.pie()
		.value(
			function(d) {
				return d.value;
			}
		)
		.sort(
			null
		)
	;
	var arcs = Container.selectAll("g.slice")
		.data(pie)
		.enter()
			.append("svg:g")
				.attr("class", "slice")
	;
	//Create the drawn data for each pie slice
	arcs.append("svg:path")
		.attr("fill", 
			function(d, i) {
				return color(i);
			}
		)
		.attr("d", 
			//Set the radius of each slice, to indicate the 'fullness' of the class it represents
			d3.svg.arc().outerRadius(
				function() {
					return radius;
				}
			)
		)
		.append("svg:title")
			.text(
				function(d,i) { return data[i].value + " " + data[i].id;
			})

	;

	//Add text labels to each pie slice if needed
	arcs.append("svg:text")
		.attr("class","tinytext")
		.attr("transform", 
			function(d) {
				var rotation = 
					(
						(
							(d.startAngle + d.endAngle)
							/
							(4*Math.PI)
						) -.25
					) * 360
				;
				if (rotation > 90 && rotation < 270) {
					return "rotate(" 
						+ 
						(360 - (180 - rotation))
						+ ") translate(" + (-radius - 4) + ",0)"
					;
				}
				else {
					return "rotate(" 
						+ 
						rotation
						+ ") translate(" + (radius + 4) + ",0)"
					;
				}
			}
		)
		.attr("text-anchor", function(d) {
				var rotation = 
					(
						(
							(d.startAngle + d.endAngle)
							/
							(4*Math.PI)
						) -.25
					) * 360
				;
				if (rotation > 90 && rotation < 270) return "end";
				
				else return "start";
				
			}
		)
		.text(
			function(d,i) { 
				return data[i].id;
			}
		)
	;
}


function mouseUp() {
	var div = document.getElementById('tooltip');
	if (!div) return;
	Sankey_TooltipInitialPos.x = Number(div.style.left.substr(0,div.style.left.length-2));
	Sankey_TooltipInitialPos.y = Number(div.style.top.substr(0,div.style.top.length-2));
    window.removeEventListener('mousemove', tooltipDrag, true);
}

function mouseDown(e) {
	Sankey_TooltipStartDrag.x = e.clientX + window.scrollX;
	Sankey_TooltipStartDrag.y = e.clientY + window.scrollY;
	window.addEventListener('mousemove', tooltipDrag, true);
}

function tooltipDrag(e){
    var div = document.getElementById('tooltip');
	div.style.position = 'absolute';
	div.style.top = (Sankey_TooltipInitialPos.y + (e.clientY - (Sankey_TooltipStartDrag.y - window.scrollY)))  + 'px';
	div.style.left = (Sankey_TooltipInitialPos.x + (e.clientX - (Sankey_TooltipStartDrag.x - window.scrollX)))  + 'px';
}
function getScrollBarWidth () {
	var inner = document.createElement('p');
	inner.style.width = "100%";
	inner.style.height = "200px";

	var outer = document.createElement('div');
	outer.style.position = "absolute";
	outer.style.top = "0px";
	outer.style.left = "0px";
	outer.style.visibility = "hidden";
	outer.style.width = "200px";
	outer.style.height = "150px";
	outer.style.overflow = "hidden";
	outer.appendChild (inner);

	document.body.appendChild (outer);
	var w1 = inner.offsetWidth;
	outer.style.overflow = 'scroll';
	var w2 = inner.offsetWidth;
	if (w1 == w2) w2 = outer.clientWidth;

	document.body.removeChild (outer);

	return (w1 - w2);
};
function AddStudentSearch (e) {
	if (e.which == 13 || e.keyCode == 13) {
		TryAddStudent();
		return;
	}

	/*
	if (SIDs.indexOf(input) != -1)
		d.style["color"] = "#00FF00";
	else d.style["color"] = "#FF0000";
	*/

	UpdateSearchTable();
}
function TryAddStudent() {
	var d = document.getElementById('AddStudentInput');
	var input = d.value;
	var searchfor = input.toLowerCase();
	var success = false;

	if (SIDs.indexOf(input) != -1) {
		Sankey_SelectedStudents.add(input);
		success = true;
	}
	else {
		Sankey_rawSIDs.forEach(function(d) {
			if (Sankey_rawStudents[d].name.toLowerCase() == searchfor) {
				Sankey_SelectedStudents.add(d);
				success = true;
				return;
			}
		});
	}
	if (!success)
		console.log("invalid");

	GenerateSankey();
}

function UpdateSearchTable() {
	var table = d3.select("#searchtable");
	table.selectAll('tr').remove();

	var HoverColor = "#DDDDEE";
	var AltColor = "#F0F0FF";
	var SelectColor = "#99AAFF";

	var searchtableR1 = table.append("tr").style("background-color","#FFFFFF");
	searchtableR1.append('td').html("Student ID").attr("class","SubTitle").data(["id"]);
	searchtableR1.append('td').html("Name").attr("class","SubTitle").data(["name"]);


	var d = document.getElementById('AddStudentInput');
	var input = d.value;

	if (input == "") return;

	var searchfor = input.toLowerCase();

	Sankey_LastSearchString = input;

	var SearchResults = [];

	//Are we searching by id number?
	if (searchfor.match(/^[0-9]+$/) != null) {
		Sankey_rawSIDs.forEach(function(d) {
			if (d.indexOf(searchfor) != -1)
				SearchResults.push(d);
		});
	}
	else {	//Or by name?
		Sankey_rawSIDs.forEach(function(d) {
			if (Sankey_rawStudents[d].name.toLowerCase().indexOf(searchfor) != -1)
				SearchResults.push(d);
		});
	}


	SearchResults.forEach(function(d,i) {
		var NewRow = table.append("tr");

		NewRow.data([{"d":d,"i":i}])
			.style("background-color",function(d) {
				if (Sankey_SelectedStudents.has(d.d))
					return SelectColor;

				if (d.i % 2) return AltColor;

				return "#FFFFFF";
			})
		;

		NewRow
			.on("click",function(d) {

				if (Sankey_SelectedStudents.has(d.d)) {
					if (d.i % 2) this.style["background-color"] = AltColor;
					else this.style["background-color"] = "#FFFFFF";
					Sankey_SelectedStudents.delete(d.d);
				}
				else {
					this.style["background-color"] = SelectColor;
					Sankey_SelectedStudents.add(d.d);
				}
				
				GenerateSankey();

				Tooltip();
			})
			.on("mouseover",function(d) {
				this.style["background-color"] = HoverColor;
			})
			.on("mouseout",function(d) {
				if (Sankey_SelectedStudents.has(d.d)) 
					this.style["background-color"] = SelectColor;
				else {
					if (d.i % 2) this.style["background-color"] = AltColor;
					else this.style["background-color"] = "#FFFFFF";
				}
			})
		;

		NewRow.append('td').html(d);

		NewRow.append('td').html(Sankey_rawStudents[d].name);

	});
}